from chromadb.api.types import EmbeddingFunction, Space, Embeddings, Documents
from chromadb.utils.embedding_functions.schemas import validate_config_schema
from typing import List, Dict, Any, Optional
import os
import numpy as np
import warnings


class VoyageAIEmbeddingFunction(EmbeddingFunction[Documents]):
    """
    This class is used to generate embeddings for a list of texts using the VoyageAI API.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        model_name: str = "voyage-large-2",
        api_key_env_var: str = "CHROMA_VOYAGE_API_KEY",
        input_type: Optional[str] = None,
        truncation: bool = True,
    ):
        """
        Initialize the VoyageAIEmbeddingFunction.

        Args:
            api_key_env_var (str, optional): Environment variable name that contains your API key for the VoyageAI API.
                Defaults to "CHROMA_VOYAGE_API_KEY".
            model_name (str, optional): The name of the model to use for text embeddings.
                Defaults to "voyage-large-2".
            api_key (str, optional): API key for the VoyageAI API. If not provided, will look for it in the environment variable.
            input_type (str, optional): The type of input to use for the VoyageAI API.
                Defaults to None.
            truncation (bool): Whether to truncate the input text.
                Defaults to True.
        """
        try:
            import voyageai
        except ImportError:
            raise ValueError(
                "The voyageai python package is not installed. Please install it with `pip install voyageai`"
            )

        if api_key is not None:
            warnings.warn(
                "Direct api_key configuration will not be persisted. "
                "Please use environment variables via api_key_env_var for persistent storage.",
                DeprecationWarning,
            )

        if os.getenv("VOYAGE_API_KEY") is not None:
            self.api_key_env_var = "VOYAGE_API_KEY"
        else:
            self.api_key_env_var = api_key_env_var

        self.api_key = api_key or os.getenv(self.api_key_env_var)
        if not self.api_key:
            raise ValueError(
                f"The {self.api_key_env_var} environment variable is not set."
            )

        self.model_name = model_name
        self.input_type = input_type
        self.truncation = truncation
        self._client = voyageai.Client(api_key=self.api_key)

    def __call__(self, input: Documents) -> Embeddings:
        """
        Generate embeddings for the given documents.

        Args:
            input: Documents to generate embeddings for.

        Returns:
            Embeddings for the documents.
        """
        embeddings = self._client.embed(
            texts=input,
            model=self.model_name,
            input_type=self.input_type,
            truncation=self.truncation,
        )

        # Convert to numpy arrays
        return [
            np.array(embedding, dtype=np.float32) for embedding in embeddings.embeddings
        ]

    @staticmethod
    def name() -> str:
        return "voyageai"

    def default_space(self) -> Space:
        return "cosine"

    def supported_spaces(self) -> List[Space]:
        return ["cosine", "l2", "ip"]

    @staticmethod
    def build_from_config(config: Dict[str, Any]) -> "EmbeddingFunction[Documents]":
        api_key_env_var = config.get("api_key_env_var")
        model_name = config.get("model_name")
        input_type = config.get("input_type")
        truncation = config.get("truncation")

        if api_key_env_var is None or model_name is None:
            assert False, "This code should not be reached"

        return VoyageAIEmbeddingFunction(
            api_key_env_var=api_key_env_var,
            model_name=model_name,
            input_type=input_type,
            truncation=truncation if truncation is not None else True,
        )

    def get_config(self) -> Dict[str, Any]:
        return {
            "api_key_env_var": self.api_key_env_var,
            "model_name": self.model_name,
            "input_type": self.input_type,
            "truncation": self.truncation,
        }

    def validate_config_update(
        self, old_config: Dict[str, Any], new_config: Dict[str, Any]
    ) -> None:
        if "model_name" in new_config:
            raise ValueError(
                "The model name cannot be changed after the embedding function has been initialized."
            )

    @staticmethod
    def validate_config(config: Dict[str, Any]) -> None:
        """
        Validate the configuration using the JSON schema.

        Args:
            config: Configuration to validate

        Raises:
            ValidationError: If the configuration does not match the schema
        """
        validate_config_schema(config, "voyageai")



from typing import Dict, Optional, Union
import logging
from chromadb.api.client import Client as ClientCreator
from chromadb.api.client import (
    AdminClient as AdminClientCreator,
)
from chromadb.api.async_client import AsyncClient as AsyncClientCreator
from chromadb.auth.token_authn import TokenTransportHeader
import chromadb.config
from chromadb.config import DEFAULT_DATABASE, DEFAULT_TENANT, Settings
from chromadb.api import AdminAPI, AsyncClientAPI, ClientAPI
from chromadb.api.models.Collection import Collection
from chromadb.api.types import (
    Cmek,
    CollectionMetadata,
    UpdateMetadata,
    Documents,
    EmbeddingFunction,
    Embeddings,
    URI,
    URIs,
    IDs,
    Include,
    Metadata,
    Metadatas,
    ReadLevel,
    Where,
    QueryResult,
    GetResult,
    WhereDocument,
    UpdateCollectionMetadata,
    SparseVector,
    SparseVectors,
    SparseEmbeddingFunction,
    Schema,
    VectorIndexConfig,
    HnswIndexConfig,
    SpannIndexConfig,
    FtsIndexConfig,
    SparseVectorIndexConfig,
    StringInvertedIndexConfig,
    IntInvertedIndexConfig,
    FloatInvertedIndexConfig,
    BoolInvertedIndexConfig,
)

# Import Search API components
from chromadb.execution.expression.plan import Search
from chromadb.execution.expression.operator import (
    # Key builder for where conditions and field selection
    Key,
    K,  # Alias for Key
    # KNN-based ranking for hybrid search
    Knn,
    # Reciprocal Rank Fusion for combining rankings
    Rrf,
)
from pathlib import Path
import os

# Re-export types from chromadb.types
__all__ = [
    "Cmek",
    "Collection",
    "Metadata",
    "Metadatas",
    "Where",
    "WhereDocument",
    "Documents",
    "IDs",
    "URI",
    "URIs",
    "Embeddings",
    "EmbeddingFunction",
    "Include",
    "CollectionMetadata",
    "UpdateMetadata",
    "UpdateCollectionMetadata",
    "QueryResult",
    "GetResult",
    "TokenTransportHeader",
    # Search API components
    "Search",
    "Key",
    "K",
    "Knn",
    "Rrf",
    # Sparse Vector Types
    "SparseVector",
    "SparseVectors",
    "SparseEmbeddingFunction",
    # Schema and Index Configuration
    "Schema",
    "VectorIndexConfig",
    "HnswIndexConfig",
    "SpannIndexConfig",
    "FtsIndexConfig",
    "SparseVectorIndexConfig",
    "StringInvertedIndexConfig",
    "IntInvertedIndexConfig",
    "FloatInvertedIndexConfig",
    "BoolInvertedIndexConfig",
]

from chromadb.types import CloudClientArg

logger = logging.getLogger(__name__)

__settings = Settings()

__version__ = "1.5.1"


# Workaround to deal with Colab's old sqlite3 version
def is_in_colab() -> bool:
    try:
        import google.colab  # noqa: F401

        return True
    except ImportError:
        return False


IN_COLAB = is_in_colab()

is_client = False
try:
    from chromadb.is_thin_client import is_thin_client

    is_client = is_thin_client
except ImportError:
    is_client = False

if not is_client:
    import sqlite3

    if sqlite3.sqlite_version_info < (3, 35, 0):
        if IN_COLAB:
            # In Colab, hotswap to pysqlite-binary if it's too old
            import subprocess
            import sys

            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", "pysqlite3-binary"]
            )
            __import__("pysqlite3")
            sys.modules["sqlite3"] = sys.modules.pop("pysqlite3")
        else:
            raise RuntimeError(
                "\033[91mYour system has an unsupported version of sqlite3. Chroma \
                    requires sqlite3 >= 3.35.0.\033[0m\n"
                "\033[94mPlease visit \
                    https://docs.trychroma.com/troubleshooting#sqlite to learn how \
                    to upgrade.\033[0m"
            )


def configure(**kwargs) -> None:  # type: ignore
    """Override Chroma's default settings, environment variables or .env files"""
    global __settings
    __settings = chromadb.config.Settings(**kwargs)


def get_settings() -> Settings:
    return __settings


def EphemeralClient(
    settings: Optional[Settings] = None,
    tenant: str = DEFAULT_TENANT,
    database: str = DEFAULT_DATABASE,
) -> ClientAPI:
    """Create an in-memory client for local use.

    This client stores all data in memory and does not persist to disk.
    It is intended for testing and development.

    Args:
        settings: Optional settings to override defaults.
        tenant: Tenant name to use for requests. Defaults to the default tenant.
        database: Database name to use for requests. Defaults to the default database.

    Returns:
        ClientAPI: A configured client instance.
    """
    if settings is None:
        settings = Settings()
    settings.is_persistent = False

    # Make sure paramaters are the correct types -- users can pass anything.
    tenant = str(tenant)
    database = str(database)

    return ClientCreator(settings=settings, tenant=tenant, database=database)


def PersistentClient(
    path: Union[str, Path] = "./chroma",
    settings: Optional[Settings] = None,
    tenant: str = DEFAULT_TENANT,
    database: str = DEFAULT_DATABASE,
) -> ClientAPI:
    """Create a persistent client that stores data on disk.

    This client is intended for local development and testing. For production,
    prefer a server-backed Chroma instance.

    Args:
        path: Directory to store persisted data.
        settings: Optional settings to override defaults.
        tenant: Tenant name to use for requests.
        database: Database name to use for requests.

    Returns:
        ClientAPI: A configured client instance.
    """
    if settings is None:
        settings = Settings()
    settings.persist_directory = str(path)
    settings.is_persistent = True

    # Make sure paramaters are the correct types -- users can pass anything.
    tenant = str(tenant)
    database = str(database)

    return ClientCreator(tenant=tenant, database=database, settings=settings)


def RustClient(
    path: Optional[str] = None,
    settings: Optional[Settings] = None,
    tenant: str = DEFAULT_TENANT,
    database: str = DEFAULT_DATABASE,
) -> ClientAPI:
    """
    Creates an ephemeral or persistance instance of Chroma that saves to disk.
    This is useful for testing and development, but not recommended for production use.

    Args:
        path: An optional directory to save Chroma's data to. The client is ephemeral if a None value is provided. Defaults to None.
        tenant: The tenant to use for this client.
        database: The database to use for this client.
    """
    if settings is None:
        settings = Settings()

    settings.chroma_api_impl = "chromadb.api.rust.RustBindingsAPI"
    settings.is_persistent = path is not None
    settings.persist_directory = path or ""

    # Make sure paramaters are the correct types -- users can pass anything.
    tenant = str(tenant)
    database = str(database)

    return ClientCreator(tenant=tenant, database=database, settings=settings)


def HttpClient(
    host: str = "localhost",
    port: int = 8000,
    ssl: bool = False,
    headers: Optional[Dict[str, str]] = None,
    settings: Optional[Settings] = None,
    tenant: str = DEFAULT_TENANT,
    database: str = DEFAULT_DATABASE,
) -> ClientAPI:
    """Create a client that connects to a Chroma server.

    Args:
        host: Hostname of the Chroma server.
        port: HTTP port of the Chroma server.
        ssl: Whether to enable SSL for the connection.
        headers: Optional headers to send with each request.
        settings: Optional settings to override defaults.
        tenant: Tenant name to use for requests.
        database: Database name to use for requests.

    Returns:
        ClientAPI: A configured client instance.

    Raises:
        ValueError: If settings specify a different host or port.
    """

    if settings is None:
        settings = Settings()

    # Make sure parameters are the correct types -- users can pass anything.
    host = str(host)
    port = int(port)
    ssl = bool(ssl)
    tenant = str(tenant)
    database = str(database)

    settings.chroma_api_impl = "chromadb.api.fastapi.FastAPI"
    if settings.chroma_server_host and settings.chroma_server_host != host:
        raise ValueError(
            f"Chroma server host provided in settings[{settings.chroma_server_host}] is different to the one provided in HttpClient: [{host}]"
        )
    settings.chroma_server_host = host
    if settings.chroma_server_http_port and settings.chroma_server_http_port != port:
        raise ValueError(
            f"Chroma server http port provided in settings[{settings.chroma_server_http_port}] is different to the one provided in HttpClient: [{port}]"
        )
    settings.chroma_server_http_port = port
    settings.chroma_server_ssl_enabled = ssl
    settings.chroma_server_headers = headers

    return ClientCreator(tenant=tenant, database=database, settings=settings)


async def AsyncHttpClient(
    host: str = "localhost",
    port: int = 8000,
    ssl: bool = False,
    headers: Optional[Dict[str, str]] = None,
    settings: Optional[Settings] = None,
    tenant: str = DEFAULT_TENANT,
    database: str = DEFAULT_DATABASE,
) -> AsyncClientAPI:
    """Create an async client that connects to a Chroma HTTP server.

    This supports multiple clients connecting to the same server and is the
    recommended production configuration.

    Args:
        host: Hostname of the Chroma server.
        port: HTTP port of the Chroma server.
        ssl: Whether to enable SSL for the connection.
        headers: Optional headers to send with each request.
        settings: Optional settings to override defaults.
        tenant: Tenant name to use for requests.
        database: Database name to use for requests.

    Returns:
        AsyncClientAPI: A configured async client instance.

    Raises:
        ValueError: If settings specify a different host or port.
    """

    if settings is None:
        settings = Settings()

    # Make sure parameters are the correct types -- users can pass anything.
    host = str(host)
    port = int(port)
    ssl = bool(ssl)
    tenant = str(tenant)
    database = str(database)

    settings.chroma_api_impl = "chromadb.api.async_fastapi.AsyncFastAPI"
    if settings.chroma_server_host and settings.chroma_server_host != host:
        raise ValueError(
            f"Chroma server host provided in settings[{settings.chroma_server_host}] is different to the one provided in HttpClient: [{host}]"
        )
    settings.chroma_server_host = host
    if settings.chroma_server_http_port and settings.chroma_server_http_port != port:
        raise ValueError(
            f"Chroma server http port provided in settings[{settings.chroma_server_http_port}] is different to the one provided in HttpClient: [{port}]"
        )
    settings.chroma_server_http_port = port
    settings.chroma_server_ssl_enabled = ssl
    settings.chroma_server_headers = headers

    return await AsyncClientCreator.create(
        tenant=tenant, database=database, settings=settings
    )


def CloudClient(
    tenant: Optional[str] = None,
    database: Optional[str] = None,
    api_key: Optional[str] = None,
    settings: Optional[Settings] = None,
    *,  # Following arguments are keyword-only, intended for testing only.
    cloud_host: str = "api.trychroma.com",
    cloud_port: int = 443,
    enable_ssl: bool = True,
) -> ClientAPI:
    """Create a client for Chroma Cloud.

    If not provided, `tenant`, `database`, and `api_key` will be inferred from the environment variables `CHROMA_TENANT`, `CHROMA_DATABASE`, and `CHROMA_API_KEY`.

    Args:
        tenant: Tenant name to use, or None to infer from credentials.
        database: Database name to use, or None to infer from credentials.
        api_key: API key for Chroma Cloud.
        settings: Optional settings to override defaults.

    Returns:
        ClientAPI: A configured client instance.

    Raises:
        ValueError: If no API key is provided or available in the environment.
    """

    required_args = [
        CloudClientArg(name="api_key", env_var="CHROMA_API_KEY", value=api_key),
    ]

    # If api_key is not provided, try to load it from the environment variable
    if not all([arg.value for arg in required_args]):
        for arg in required_args:
            arg.value = arg.value or os.environ.get(arg.env_var)

    missing_args = [arg for arg in required_args if arg.value is None]
    if missing_args:
        raise ValueError(
            f"Missing required arguments: {', '.join([arg.name for arg in missing_args])}. "
            f"Please provide them or set the environment variables: {', '.join([arg.env_var for arg in missing_args])}"
        )

    if settings is None:
        settings = Settings()

    # Make sure paramaters are the correct types -- users can pass anything.
    tenant = tenant or os.environ.get("CHROMA_TENANT")
    if tenant is not None:
        tenant = str(tenant)
    database = database or os.environ.get("CHROMA_DATABASE")
    if database is not None:
        database = str(database)
    api_key = str(api_key)
    cloud_host = str(cloud_host)
    cloud_port = int(cloud_port)
    enable_ssl = bool(enable_ssl)

    settings.chroma_api_impl = "chromadb.api.fastapi.FastAPI"
    settings.chroma_server_host = cloud_host
    settings.chroma_server_http_port = cloud_port
    settings.chroma_server_ssl_enabled = enable_ssl

    settings.chroma_client_auth_provider = (
        "chromadb.auth.token_authn.TokenAuthClientProvider"
    )
    settings.chroma_client_auth_credentials = api_key
    settings.chroma_auth_token_transport_header = TokenTransportHeader.X_CHROMA_TOKEN
    settings.chroma_overwrite_singleton_tenant_database_access_from_auth = True

    return ClientCreator(tenant=tenant, database=database, settings=settings)


def Client(
    settings: Settings = __settings,
    tenant: str = DEFAULT_TENANT,
    database: str = DEFAULT_DATABASE,
) -> ClientAPI:
    """
    Return a running chroma.API instance

    tenant: The tenant to use for this client. Defaults to the default tenant.
    database: The database to use for this client. Defaults to the default database.
    """

    # Make sure paramaters are the correct types -- users can pass anything.
    tenant = str(tenant)
    database = str(database)

    return ClientCreator(tenant=tenant, database=database, settings=settings)


def AdminClient(settings: Settings = Settings()) -> AdminAPI:
    """Create an admin client for tenant and database management."""
    return AdminClientCreator(settings=settings)



import gc
import math
import os.path
from uuid import UUID
from contextlib import contextmanager

from chromadb.api.segment import SegmentAPI
from chromadb.db.system import SysDB
from chromadb.ingest.impl.utils import create_topic_name

from chromadb.config import System
from chromadb.db.base import get_sql
from chromadb.db.impl.sqlite import SqliteDB
from time import sleep
import psutil

from chromadb.segment import SegmentType
from chromadb.test.property.strategies import NormalizedRecordSet, RecordSet
from typing import Callable, Optional, Tuple, Union, List, TypeVar, cast, Any, Dict
from typing_extensions import Literal
import numpy as np
import numpy.typing as npt
from chromadb.api import types, ClientAPI
from chromadb.api.models.Collection import Collection
from hypothesis import note
from hypothesis.errors import InvalidArgument
from pypika import Table, functions

from chromadb.utils import distance_functions
from chromadb.execution.expression.plan import Search
from chromadb.execution.expression.operator import Knn, Select, Limit, Key

T = TypeVar("T")


def wrap(value: Union[T, List[T]]) -> List[T]:
    """Wrap a value in a list if it is not a list"""
    if value is None:
        raise InvalidArgument("value cannot be None")
    elif isinstance(value, List):
        return value
    else:
        return [value]


def wrap_all(record_set: RecordSet) -> NormalizedRecordSet:
    """Ensure that an embedding set has lists for all its values"""

    embedding_list: Optional[types.Embeddings]
    if record_set["embeddings"] is None:
        embedding_list = None
    elif isinstance(record_set["embeddings"], list):
        assert record_set["embeddings"] is not None
        if len(record_set["embeddings"]) > 0:
            if all(
                isinstance(embedding, list) for embedding in record_set["embeddings"]
            ):
                embedding_list = cast(types.Embeddings, record_set["embeddings"])
            elif all(
                isinstance(embedding, np.ndarray)
                for embedding in record_set["embeddings"]
            ):
                embedding_list = cast(types.Embeddings, record_set["embeddings"])
            else:
                if all(
                    isinstance(e, (int, float, np.integer, np.floating))
                    for e in record_set["embeddings"]
                ):
                    embedding_list = cast(types.Embeddings, [record_set["embeddings"]])
                else:
                    raise InvalidArgument(
                        "an embedding must be a list of floats or ints"
                    )
        else:
            embedding_list = cast(types.Embeddings, record_set["embeddings"])
    else:
        raise InvalidArgument(
            "embeddings must be a list of lists, a list of numpy arrays, a list of numbers, or None"
        )

    return NormalizedRecordSet(
        ids=wrap(record_set["ids"]),
        documents=wrap(record_set["documents"])
        if record_set["documents"] is not None
        else None,
        metadatas=wrap(record_set["metadatas"])
        if record_set["metadatas"] is not None
        else None,
        embeddings=embedding_list,
    )


def check_metadata(
    expected: Optional[types.Metadata], got: Optional[types.Metadata]
) -> None:
    assert (expected is None and got is None) or (
        expected is not None and got is not None
    )
    if expected is not None and got is not None:
        assert len(expected) == len(got)
        for key, val in expected.items():
            assert key in got
            if isinstance(expected[key], float) and isinstance(got[key], float):
                assert abs(cast(float, expected[key]) - cast(float, got[key])) < 1e-6
            else:
                assert expected[key] == got[key]


def count(collection: Collection, record_set: RecordSet) -> None:
    """The given collection count is equal to the number of embeddings"""
    count = collection.count()
    normalized_record_set = wrap_all(record_set)
    if count != len(normalized_record_set["ids"]):
        print("count mismatch:", count, "=!", len(normalized_record_set["ids"]))
    assert count == len(normalized_record_set["ids"])


def _field_matches(
    collection: Collection,
    normalized_record_set: NormalizedRecordSet,
    field_name: Union[
        Literal["documents"], Literal["metadatas"], Literal["embeddings"]
    ],
) -> None:
    """
    The actual embedding field is equal to the expected field
    field_name: one of [documents, metadatas]
    """
    result = collection.get(ids=normalized_record_set["ids"], include=[field_name])  # type: ignore[list-item]
    # The test_out_of_order_ids test fails because of this in test_add.py
    # Here we sort by the ids to match the input order
    embedding_id_to_index = {id: i for i, id in enumerate(normalized_record_set["ids"])}
    actual_field = result[field_name]

    if len(normalized_record_set["ids"]) == 0:
        if field_name == "embeddings":
            assert cast(npt.NDArray[Any], actual_field).size == 0
        else:
            assert actual_field == []
        return

    # This assert should never happen, if we include metadatas/documents it will be
    # [None, None..] if there is no metadata. It will not be just None.
    assert actual_field is not None
    sorted_field = sorted(
        enumerate(actual_field),
        key=lambda index_and_field_value: embedding_id_to_index[
            result["ids"][index_and_field_value[0]]
        ],
    )
    field_values = [field_value for _, field_value in sorted_field]

    expected_field = normalized_record_set[field_name]
    if expected_field is None:
        # Since an RecordSet is the user input, we need to convert the documents to
        # a List since thats what the API returns -> none per entry
        expected_field = [None] * len(normalized_record_set["ids"])  # type: ignore
    if field_name == "embeddings":
        assert np.allclose(np.array(field_values), np.array(expected_field))
    else:
        assert len(field_values) == len(expected_field)

        for field_value, expected_field in zip(field_values, expected_field):
            if isinstance(expected_field, dict):
                check_metadata(
                    cast(types.Metadata, field_value),
                    cast(types.Metadata, expected_field),
                )
            else:
                assert field_value == expected_field


def ids_match(collection: Collection, record_set: RecordSet) -> None:
    """The actual embedding ids is equal to the expected ids"""
    normalized_record_set = wrap_all(record_set)
    actual_ids = collection.get(ids=normalized_record_set["ids"], include=[])["ids"]
    # The test_out_of_order_ids test fails because of this in test_add.py
    # Here we sort the ids to match the input order
    embedding_id_to_index = {id: i for i, id in enumerate(normalized_record_set["ids"])}
    actual_ids = sorted(actual_ids, key=lambda id: embedding_id_to_index[id])
    assert actual_ids == normalized_record_set["ids"]


def metadatas_match(collection: Collection, record_set: RecordSet) -> None:
    """The actual embedding metadata is equal to the expected metadata"""
    normalized_record_set = wrap_all(record_set)
    _field_matches(collection, normalized_record_set, "metadatas")


def documents_match(collection: Collection, record_set: RecordSet) -> None:
    """The actual embedding documents is equal to the expected documents"""
    normalized_record_set = wrap_all(record_set)
    _field_matches(collection, normalized_record_set, "documents")


def embeddings_match(collection: Collection, record_set: RecordSet) -> None:
    """The actual embedding documents is equal to the expected documents"""
    normalized_record_set = wrap_all(record_set)
    _field_matches(collection, normalized_record_set, "embeddings")


def no_duplicates(collection: Collection) -> None:
    ids = collection.get()["ids"]
    assert len(ids) == len(set(ids))


def _exact_distances(
    query: types.Embeddings,
    targets: types.Embeddings,
    distance_fn: Callable[
        [npt.ArrayLike, npt.ArrayLike], float
    ] = distance_functions.l2,
) -> Tuple[List[List[int]], List[List[float]]]:
    """Return the ordered indices and distances from each query to each target"""
    np_query = np.array(query, dtype=np.float32)
    np_targets = np.array(targets, dtype=np.float32)

    # Compute the distance between each query and each target, using the distance function
    distances = np.apply_along_axis(
        lambda query: np.apply_along_axis(distance_fn, 1, np_targets, query),
        1,
        np_query,
    )
    # Sort the distances and return the indices
    return np.argsort(distances).tolist(), distances.tolist()


def fd_not_exceeding_threadpool_size(threadpool_size: int) -> None:
    """
    Checks that the open file descriptors are not exceeding the threadpool size
    works only for SegmentAPI
    """
    current_process = psutil.Process()
    open_files = current_process.open_files()
    max_retries = 5
    retry_count = 0
    # we probably don't need the below but we keep it to avoid flaky tests.
    while (
        len([p.path for p in open_files if "sqlite3" in p.path]) - 1 > threadpool_size
        and retry_count < max_retries
    ):
        gc.collect()  # GC to collect the orphaned TLS objects
        open_files = current_process.open_files()
        retry_count += 1
        sleep(1)
    assert (
        len([p.path for p in open_files if "sqlite3" in p.path]) - 1 <= threadpool_size
    )


def get_space(collection: Collection):
    # TODO: this is a hack to get the space
    # We should update the tests to not pass space via metadata instead use collection
    # configuration_json
    space = None
    if "hnsw:space" in collection.metadata:
        space = collection.metadata["hnsw:space"]
    if collection._model.configuration_json is None:
        return space
    if (
        "spann" in collection._model.configuration_json
        and collection._model.configuration_json.get("spann") is not None
        and "space" in collection._model.configuration_json.get("spann")
    ):
        space = collection._model.configuration_json.get("spann").get("space")
    elif (
        "hnsw" in collection._model.configuration_json
        and collection._model.configuration_json.get("hnsw") is not None
        and "space" in collection._model.configuration_json.get("hnsw")
    ):
        if space is None:
            space = collection._model.configuration_json.get("hnsw").get("space")
    return space


def ann_accuracy(
    collection: Collection,
    record_set: RecordSet,
    n_results: int = 1,
    min_recall: float = 0.95,
    embedding_function: Optional[types.EmbeddingFunction] = None,  # type: ignore[type-arg]
    query_indices: Optional[List[int]] = None,
    query_embeddings: Optional[types.Embeddings] = None,
    use_search: bool = False,
) -> None:
    """Validate that the API performs nearest_neighbor searches correctly"""
    normalized_record_set = wrap_all(record_set)

    if len(normalized_record_set["ids"]) == 0:
        return  # nothing to test here

    embeddings: Optional[types.Embeddings] = normalized_record_set["embeddings"]
    have_embeddings = embeddings is not None and len(embeddings) > 0
    if not have_embeddings:
        assert embedding_function is not None
        assert normalized_record_set["documents"] is not None
        assert isinstance(normalized_record_set["documents"], list)
        # Compute the embeddings for the documents
        embeddings = embedding_function(normalized_record_set["documents"])

    space = get_space(collection)
    if space is None:
        distance_function = distance_functions.l2
    elif space == "cosine":
        distance_function = distance_functions.cosine
    elif space == "ip":
        distance_function = distance_functions.ip
    elif space == "l2":
        distance_function = distance_functions.l2

    accuracy_threshold = 1e-6
    assert collection.metadata is not None
    assert embeddings is not None
    # TODO: ip and cosine are numerically unstable in HNSW.
    # The higher the dimensionality, the more noise is introduced, since each float element
    # of the vector has noise added, which is then subsequently included in all normalization calculations.
    # This means that higher dimensions will have more noise, and thus more error.
    assert all(isinstance(e, (list, np.ndarray)) for e in embeddings)
    dim = len(embeddings[0])
    accuracy_threshold = accuracy_threshold * math.pow(10, int(math.log10(dim)))

    # Perform exact distance computation
    if query_embeddings is None:
        query_embeddings = (
            embeddings
            if query_indices is None
            else [embeddings[i] for i in query_indices]
        )
    query_documents = normalized_record_set["documents"]
    if query_indices is not None and query_documents is not None:
        query_documents = [query_documents[i] for i in query_indices]

    indices, distances = _exact_distances(
        query_embeddings, embeddings, distance_fn=distance_function
    )

    if use_search:
        # Use search API instead of query
        search_requests = []
        for query_embedding in query_embeddings:
            # Convert numpy array to list if needed
            if isinstance(query_embedding, np.ndarray):
                query_embedding_list = query_embedding.tolist()
            else:
                query_embedding_list = query_embedding
            search = Search(
                rank=Knn(query=query_embedding_list),
                limit=Limit(limit=n_results),
            ).select_all()
            search_requests.append(search)

        # Call search API
        search_results = collection.search(
            searches=search_requests,
        )

        # Convert search results to query-like format
        query_results = cast(
            types.QueryResult,
            {
                "ids": search_results["ids"],
                "distances": search_results[
                    "scores"
                ],  # scores is distances in search API
                "embeddings": search_results["embeddings"],
                "documents": search_results["documents"],
                "metadatas": search_results["metadatas"],
            },
        )
    else:
        query_results = collection.query(
            query_embeddings=query_embeddings if have_embeddings else None,
            query_texts=query_documents if not have_embeddings else None,
            n_results=n_results,
            include=["embeddings", "documents", "metadatas", "distances"],  # type: ignore[list-item]
        )

    _query_results_are_correct_shape(query_results, n_results)

    # Assert fields are not None for type checking
    assert query_results["ids"] is not None
    assert query_results["distances"] is not None
    assert query_results["embeddings"] is not None
    assert query_results["documents"] is not None
    assert query_results["metadatas"] is not None

    # Dict of ids to indices
    id_to_index = {id: i for i, id in enumerate(normalized_record_set["ids"])}
    missing = 0
    for i, (indices_i, distances_i) in enumerate(zip(indices, distances)):
        expected_ids = np.array(normalized_record_set["ids"])[indices_i[:n_results]]
        missing += len(set(expected_ids) - set(query_results["ids"][i]))

        # For each id in the query results, find the index in the embeddings set
        # and assert that the embeddings are the same
        for j, id in enumerate(query_results["ids"][i]):
            # This may be because the true nth nearest neighbor didn't get returned by the ANN query
            unexpected_id = id not in expected_ids
            index = id_to_index[id]

            correct_distance = np.allclose(
                distances_i[index],
                query_results["distances"][i][j],
                atol=accuracy_threshold,
            )
            if unexpected_id:
                # If the ID is unexpcted, but the distance is correct, then we
                # have a duplicate in the data. In this case, we should not reduce recall.
                if correct_distance:
                    missing -= 1
                else:
                    continue
            else:
                assert correct_distance

            assert np.allclose(embeddings[index], query_results["embeddings"][i][j])
            if normalized_record_set["documents"] is not None:
                assert (
                    normalized_record_set["documents"][index]
                    == query_results["documents"][i][j]
                )
            if normalized_record_set["metadatas"] is not None:
                check_metadata(
                    normalized_record_set["metadatas"][index],
                    query_results["metadatas"][i][j],
                )

    size = len(normalized_record_set["ids"])
    recall = (size - missing) / size

    try:
        note(
            f"# recall: {recall}, missing {missing} out of {size}, accuracy threshold {accuracy_threshold}"
        )
    except InvalidArgument:
        pass  # it's ok if we're running outside hypothesis

    assert recall >= min_recall

    # Ensure that the query results are sorted by distance
    for distance_result in query_results["distances"]:
        assert np.allclose(np.sort(distance_result), distance_result)


def _query_results_are_correct_shape(
    query_results: types.QueryResult, n_results: int
) -> None:
    for result_type in ["distances", "embeddings", "documents", "metadatas"]:
        assert query_results[result_type] is not None  # type: ignore[literal-required]
        assert all(
            len(result) == n_results for result in query_results[result_type]  # type: ignore[literal-required]
        )


def _total_embedding_queue_log_size(sqlite: SqliteDB) -> int:
    t = Table("embeddings_queue")
    q = sqlite.querybuilder().from_(t)

    with sqlite.tx() as cur:
        sql, params = get_sql(
            q.select(functions.Count(t.seq_id)), sqlite.parameter_format()
        )
        result = cur.execute(sql, params)
        return cast(int, result.fetchone()[0])


def log_size_below_max(
    system: System, collections: List[Collection], has_collection_mutated: bool
) -> None:
    sqlite = system.instance(SqliteDB)

    # Ephemeral Rust client is using its own sqlite impl, which cannot be accessed from Python
    if (
        not system.settings.is_persistent
        and system.settings.chroma_api_impl == "chromadb.api.rust.RustBindingsAPI"
    ):
        return

    if has_collection_mutated:
        # Must always keep one entry to avoid reusing seq_ids
        assert _total_embedding_queue_log_size(sqlite) >= 1

        # We purge per-collection as the sync_threshold is a per-collection setting
        sync_threshold_sum = sum(
            collection.metadata.get("hnsw:sync_threshold", 1000)
            if collection.metadata is not None
            else 1000
            for collection in collections
        )
        batch_size_sum = sum(
            collection.metadata.get("hnsw:batch_size", 100)
            if collection.metadata is not None
            else 100
            for collection in collections
        )

        limit = (
            sync_threshold_sum
            if system.settings.chroma_api_impl == "chromadb.api.rust.RustBindingsAPI"
            else sync_threshold_sum + batch_size_sum
        )

        # -1 is used because the queue is always at least 1 entry long, so deletion stops before the max ack'ed sequence ID.
        # And for python impl if the batch_size != sync_threshold, the queue can have up to batch_size more entries.
        assert _total_embedding_queue_log_size(sqlite) - 1 <= limit
    else:
        assert _total_embedding_queue_log_size(sqlite) == 0


def _total_embedding_queue_log_size_per_collection(
    system: System,
    collections: List[Collection],
) -> Dict[UUID, int]:
    sqlite = system.instance(SqliteDB)
    t = Table("embeddings_queue")
    q = sqlite.querybuilder().from_(t)
    _tenant = system.settings.require("tenant_id")
    _topic_namespace = system.settings.require("topic_namespace")
    topic_mappings = {
        create_topic_name(_tenant, _topic_namespace, collection.id): collection
        for collection in collections
    }
    with sqlite.tx() as cur:
        sql, params = get_sql(
            q.select(t.topic, functions.Count(t.seq_id)).groupby("topic"),
            sqlite.parameter_format(),
        )
        result = cur.execute(sql, params)
        out = {}
        for res in result.fetchall():
            out[topic_mappings[res[0]].id] = res[1]
        return out


def log_size_for_collections_match_expected(
    system: System, collections: List[Collection], has_collection_mutated: bool
) -> None:
    if system.settings.chroma_api_impl == "chromadb.api.rust.RustBindingsAPI":
        # The rust impl does not use batch size
        return

    sqlite = system.instance(SqliteDB)

    if has_collection_mutated:
        # Must always keep one entry to avoid reusing seq_ids
        assert _total_embedding_queue_log_size(sqlite) >= 1

        batch_size_sum = {
            collection.id: collection.metadata.get("hnsw:batch_size", 100)
            if collection.metadata is not None
            else 100
            for collection in collections
        }
        expected_sizes = {
            collection.id: collection.count() % batch_size_sum[collection.id] + 1
            for collection in collections
        }

        actual_sizes = _total_embedding_queue_log_size_per_collection(
            system, collections
        )
        assert set(actual_sizes.keys()) == set(expected_sizes.keys())
        assert all(
            actual_sizes[collection.id] == expected_sizes[collection.id]
            for collection in collections
        )

    else:
        assert _total_embedding_queue_log_size(sqlite) == 0


@contextmanager
def collection_deleted(client: ClientAPI, collection_name: str):
    # Invariant checks before deletion
    collection_names = [c.name for c in client.list_collections()]
    assert collection_name in collection_names
    collection = client.get_collection(collection_name)
    segments = []
    if isinstance(client._server, SegmentAPI):  # type: ignore
        sysdb: SysDB = client._server._sysdb  # type: ignore
        segments = sysdb.get_segments(collection=collection.id)
        segment_types = {}
        should_have_hnsw = False
        for segment in segments:
            segment_types[segment["type"]] = True
            if segment["type"] == SegmentType.HNSW_LOCAL_PERSISTED.value:
                sync_threshold = (
                    collection.metadata["hnsw:sync_threshold"]
                    if collection.metadata is not None
                    and "hnsw:sync_threshold" in collection.metadata
                    else 1000
                )
                if (
                    collection.count() > sync_threshold
                ):  # we only check if vector segment dir exists if we've synced at least once
                    should_have_hnsw = True
                    assert os.path.exists(
                        os.path.join(
                            client.get_settings().persist_directory, str(segment["id"])
                        )
                    )
        if should_have_hnsw:
            assert segment_types[SegmentType.HNSW_LOCAL_PERSISTED.value]
        assert segment_types[SegmentType.SQLITE.value]

    yield

    # Invariant checks after deletion
    collection_names = [c.name for c in client.list_collections()]
    assert collection_name not in collection_names
    if len(segments) > 0:
        sysdb: SysDB = client._server._sysdb  # type: ignore
        segments_after = sysdb.get_segments(collection=collection.id)
        assert len(segments_after) == 0
        for segment in segments:
            if segment["type"] == SegmentType.HNSW_LOCAL_PERSISTED.value:
                assert not os.path.exists(
                    os.path.join(
                        client.get_settings().persist_directory, str(segment["id"])
                    )
                )



from chromadb.api.types import (
    Embeddings,
    Documents,
    EmbeddingFunction,
    Space,
)
from chromadb.utils.embedding_functions.schemas import validate_config_schema
from typing import List, Dict, Any, TypedDict, Optional
import os
import numpy as np


class NomicQueryConfig(TypedDict):
    task_type: str


class NomicEmbeddingFunction(EmbeddingFunction[Documents]):
    """
    This class is used to get embeddings for a list of texts using the Nomic API.
    """

    def __init__(
        self,
        model: str,
        task_type: str,
        query_config: Optional[NomicQueryConfig],
        api_key_env_var: str = "NOMIC_API_KEY",
    ):
        """
        Initialize the NomicEmbeddingFunction.

        Args:
            model (str): The name of the model to use for text embeddings.
            task_type (str): The type of task to embed with. See reference https://docs.nomic.ai/platform/embeddings-and-retrieval/text-embedding#embedding-task-types
            query_config (Optional[NomicQueryConfig]): The configuration for setting task type for queries
            api_key_env_var (str): The environment variable name for the Nomic API key. Defaults to "NOMIC_API_KEY".

            Supported task types: search_document, search_query, classification, clustering
        """
        try:
            from nomic import embed
        except ImportError:
            raise ValueError(
                "The nomic python package is not installed. Please install it with `pip install nomic`"
            )

        self.model = model
        self.task_type = task_type
        self.api_key_env_var = api_key_env_var
        self.api_key = os.getenv(api_key_env_var)
        self.query_config = query_config
        if not self.api_key:
            raise ValueError(f"The {api_key_env_var} environment variable is not set.")
        self.embed = embed

    def __call__(self, input: Documents) -> Embeddings:
        if not all(isinstance(item, str) for item in input):
            raise ValueError("Nomic only supports text documents, not images")
        output = self.embed.text(
            model=self.model,
            texts=input,
            task_type=self.task_type,
        )
        return [np.array(data.embedding) for data in output.data]

    def embed_query(self, input: Documents) -> Embeddings:
        if not all(isinstance(item, str) for item in input):
            raise ValueError("Nomic only supports text queries, not images")

        task_type = (
            self.query_config.get("task_type") if self.query_config else self.task_type
        )
        output = self.embed.text(
            model=self.model,
            texts=input,
            task_type=task_type,
        )
        return [np.array(data.embedding) for data in output.data]

    @staticmethod
    def name() -> str:
        return "nomic"

    def default_space(self) -> Space:
        return "cosine"

    def supported_spaces(self) -> List[Space]:
        return ["cosine", "l2", "ip"]

    @staticmethod
    def build_from_config(config: Dict[str, Any]) -> "EmbeddingFunction[Documents]":
        model = config.get("model")
        api_key_env_var = config.get("api_key_env_var")
        task_type = config.get("task_type")
        query_config = config.get("query_config")
        if model is None or api_key_env_var is None or task_type is None:
            assert False, "This code should not be reached"  # this is for type checking
        return NomicEmbeddingFunction(
            model=model,
            api_key_env_var=api_key_env_var,
            task_type=task_type,
            query_config=query_config,
        )

    def get_config(self) -> Dict[str, Any]:
        return {
            "model": self.model,
            "api_key_env_var": self.api_key_env_var,
            "task_type": self.task_type,
            "query_config": self.query_config,
        }

    def validate_config_update(
        self, old_config: Dict[str, Any], new_config: Dict[str, Any]
    ) -> None:
        if "model" in new_config:
            raise ValueError(
                "The model cannot be changed after the embedding function has been initialized."
            )

    @staticmethod
    def validate_config(config: Dict[str, Any]) -> None:
        """
        Validate the configuration using the JSON schema.

        Args:
            config: Configuration to validate
        """
        validate_config_schema(config, "nomic")



import logging
from typing import Dict, Optional, Tuple
import pytest
from chromadb.api import AdminAPI
import chromadb.api.types as types
from chromadb.config import DEFAULT_DATABASE, DEFAULT_TENANT
from chromadb.test.conftest import (
    ClientFactories,
    MULTI_REGION_ENABLED,
    DEFAULT_MCMR_DATABASE,
    MULTI_REGION_TOPOLOGY,
)

from chromadb.test.property.test_collections import CollectionStateMachine
from hypothesis.stateful import (
    Bundle,
    rule,
    initialize,
    multiple,
    run_state_machine_as_test,
    MultipleResults,
)
import chromadb.test.property.strategies as strategies

EFFECTIVE_DEFAULT_DATABASE = (
    DEFAULT_MCMR_DATABASE if MULTI_REGION_ENABLED else DEFAULT_DATABASE
)


class TenantDatabaseCollectionStateMachine(CollectionStateMachine):
    """A collection state machine test that includes tenant and database information,
    and switches between them."""

    tenants: Bundle  # [str]
    databases: Bundle  # [Tuple[str, str]]  # database to tenant it belongs to
    tenant_to_database_to_model: Dict[
        str, Dict[str, Dict[str, Optional[types.CollectionMetadata]]]
    ]
    admin_client: AdminAPI
    curr_tenant: str
    curr_database: str

    tenants = Bundle("tenants")
    databases = Bundle("databases")

    def __init__(self, client_factories: ClientFactories):
        client = client_factories.create_client()
        super().__init__(client)
        self.client = client
        self.admin_client = client_factories.create_admin_client_from_system()

    @initialize()
    def initialize(self) -> None:
        self.client.reset()
        self.tenant_to_database_to_model = {}
        self.curr_tenant = DEFAULT_TENANT
        self.curr_database = EFFECTIVE_DEFAULT_DATABASE
        self.client.set_tenant(DEFAULT_TENANT, EFFECTIVE_DEFAULT_DATABASE)
        self.set_tenant_model(self.curr_tenant, {})
        self.set_database_model_for_tenant(self.curr_tenant, self.curr_database, {})

    @rule(target=tenants, name=strategies.tenant_database_name)
    def create_tenant(self, name: str) -> MultipleResults:  # [str]:
        tenant = self.overwrite_tenant(name)
        # Check if tenant already exists
        if self.has_tenant(tenant):
            with pytest.raises(Exception):
                self.admin_client.create_tenant(tenant)
            return multiple()

        self.admin_client.create_tenant(tenant)
        # When we create a tenant, create a default database for it just for testing
        # since the state machine could call collection operations before creating a
        # database
        self.admin_client.create_database(EFFECTIVE_DEFAULT_DATABASE, tenant=tenant)
        self.set_tenant_model(tenant, {})
        self.set_database_model_for_tenant(tenant, EFFECTIVE_DEFAULT_DATABASE, {})
        return multiple(tenant)

    @rule(target=databases, name=strategies.tenant_database_name)
    def create_database(self, name: str) -> MultipleResults:  # [Tuple[str, str]]:
        if MULTI_REGION_ENABLED:
            name = f"{MULTI_REGION_TOPOLOGY}+{name}"
        database = self.overwrite_database(name)
        tenant = self.overwrite_tenant(self.curr_tenant)
        # If database already exists in current tenant, raise an error
        if self.has_database_for_tenant(tenant, database):
            with pytest.raises(Exception):
                self.admin_client.create_database(name=database, tenant=tenant)
            return multiple()

        self.admin_client.create_database(name=database, tenant=tenant)
        self.set_database_model_for_tenant(
            tenant=tenant, database=database, database_model={}
        )
        return multiple((database, tenant))

    @rule(database=databases)
    def set_database_and_tenant(self, database: Tuple[str, str]) -> None:
        # Get a database and switch to the database and the tenant it belongs to
        database_name = database[0]
        tenant_name = database[1]
        self.set_api_tenant_database(tenant_name, database_name)
        self.curr_database = database_name
        self.curr_tenant = tenant_name

    @rule(tenant=tenants)
    def set_tenant(self, tenant: str) -> None:
        self.set_api_tenant_database(tenant, EFFECTIVE_DEFAULT_DATABASE)
        self.curr_tenant = tenant
        self.curr_database = EFFECTIVE_DEFAULT_DATABASE

    # These methods allow other tests, namely
    # test_collections_with_database_tenant_override.py, to swap out the model
    # without needing to do a bunch of pythonic cleverness to fake a dict which
    # preteds to have every key.
    def set_api_tenant_database(self, tenant: str, database: str) -> None:
        self.client.set_tenant(tenant, database)

    # For calls to create_database, and create_tenant we may want to override the tenant and database
    # This is a leaky abstraction that exists soley for the purpose of
    # test_collections_with_database_tenant_override.py
    def overwrite_tenant(self, tenant: str) -> str:
        return tenant

    def overwrite_database(self, database: str) -> str:
        return database

    def has_tenant(self, tenant: str) -> bool:
        return tenant in self.tenant_to_database_to_model

    def get_tenant_model(
        self, tenant: str
    ) -> Dict[str, Dict[str, Optional[types.CollectionMetadata]]]:
        return self.tenant_to_database_to_model[tenant]

    def set_tenant_model(
        self,
        tenant: str,
        model: Dict[str, Dict[str, Optional[types.CollectionMetadata]]],
    ) -> None:
        self.tenant_to_database_to_model[tenant] = model

    def has_database_for_tenant(self, tenant: str, database: str) -> bool:
        return database in self.tenant_to_database_to_model[tenant]

    def set_database_model_for_tenant(
        self,
        tenant: str,
        database: str,
        database_model: Dict[str, Optional[types.CollectionMetadata]],
    ) -> None:
        self.tenant_to_database_to_model[tenant][database] = database_model

    @property
    def model(self) -> Dict[str, Optional[types.CollectionMetadata]]:
        return self.tenant_to_database_to_model[self.curr_tenant][self.curr_database]


def test_collections(
    caplog: pytest.LogCaptureFixture, client_factories: ClientFactories
) -> None:
    caplog.set_level(logging.ERROR)
    run_state_machine_as_test(lambda: TenantDatabaseCollectionStateMachine(client_factories))  # type: ignore



from typing import Dict
from fastapi import HTTPException
from overrides import override
from chromadb.auth import (
    AuthzAction,
    AuthzResource,
    ServerAuthenticationProvider,
    ServerAuthorizationProvider,
    UserIdentity,
)
from chromadb.config import System


class ExampleAuthenticationProvider(ServerAuthenticationProvider):
    """In practice the tenant would likely be resolved from some other opaque value (e.g. key/token). Here, it's just passed directly as a header for simplicity."""

    @override
    def authenticate_or_raise(self, headers: Dict[str, str]) -> UserIdentity:
        return UserIdentity(
            user_id="test",
            tenant=headers.get("x-tenant", None),
        )


class ExampleAuthorizationProvider(ServerAuthorizationProvider):
    """A simple authz provider that asserts the user's tenant matches the resource's tenant."""

    def __init__(self, system: System) -> None:
        super().__init__(system)
        self._settings = system.settings

    @override
    def authorize_or_raise(
        self, user: UserIdentity, action: AuthzAction, resource: AuthzResource
    ) -> None:
        if user.tenant is None:
            return

        if action == AuthzAction.RESET:
            return

        if user.tenant != resource.tenant:
            raise HTTPException(status_code=403, detail="Unauthorized")



"""Utility functions for managing collection statistics.

This module provides standalone functions for enabling, disabling, and retrieving
statistics for ChromaDB collections. These functions work with the attached function
system to automatically compute metadata value frequencies.

Example:
    >>> from chromadb.utils.statistics import attach_statistics_function, get_statistics
    >>> import chromadb
    >>>
    >>> client = chromadb.Client()
    >>> collection = client.get_or_create_collection("my_collection")
    >>>
    >>> # Attach statistics function with output collection name
    >>> attach_statistics_function(collection, "my_collection_statistics")
    >>>
    >>> # Add some data
    >>> collection.add(
    ...     ids=["id1", "id2"],
    ...     documents=["doc1", "doc2"],
    ...     metadatas=[{"category": "A"}, {"category": "B"}]
    ... )
    >>>
    >>> # Get statistics from the named output collection
    >>> stats = get_statistics(collection, "my_collection_statistics")
    >>> print(stats)
"""

from typing import TYPE_CHECKING, Optional, Dict, Any, cast, Tuple
from collections import defaultdict
from chromadb.api.types import OneOrMany, Where, maybe_cast_one_to_many
from chromadb.api.functions import STATISTICS_FUNCTION

if TYPE_CHECKING:
    from chromadb.api.models.Collection import Collection
    from chromadb.api.models.AttachedFunction import AttachedFunction


def get_statistics_fn_name(collection: "Collection") -> str:
    """Generate the default name for the statistics attached function.

    Args:
        collection: The collection to generate the name for

    Returns:
        str: The statistics function name
    """
    return f"{collection.name}_stats"


def attach_statistics_function(
    collection: "Collection", stats_collection_name: str
) -> Tuple["AttachedFunction", bool]:
    """Attach statistics collection function to a collection.

    This attaches the statistics function which will automatically compute
    and update metadata value frequencies whenever records are added, updated,
    or deleted.

    Args:
        collection: The collection to enable statistics for
        stats_collection_name: Name of the collection where statistics will be stored.

    Returns:
        Tuple of (AttachedFunction, created) where created is True if newly created,
        False if already existed (idempotent request)

    Example:
        >>> attached_fn, created = attach_statistics_function(collection, "my_collection_statistics")
        >>> if created:
        ...     print("Statistics function newly attached")
        >>> collection.add(ids=["id1"], documents=["doc1"], metadatas=[{"key": "value"}])
        >>> # Statistics are automatically computed
        >>> stats = get_statistics(collection, "my_collection_statistics")
    """
    return collection.attach_function(
        function=STATISTICS_FUNCTION,
        name=get_statistics_fn_name(collection),
        output_collection=stats_collection_name,
        params=None,
    )


def get_statistics_fn(collection: "Collection") -> "AttachedFunction":
    """Get the statistics attached function for a collection.

    Args:
        collection: The collection to get the statistics function for

    Returns:
        AttachedFunction: The statistics function

    Raises:
        NotFoundError: If statistics are not enabled
        AssertionError: If the attached function is not a statistics function
    """
    af = collection.get_attached_function(get_statistics_fn_name(collection))
    assert (
        af.function_name == "statistics"
    ), "Attached function is not a statistics function"
    return af


def detach_statistics_function(
    collection: "Collection", delete_stats_collection: bool = False
) -> bool:
    """Detach statistics collection function from a collection.

    Args:
        collection: The collection to disable statistics for
        delete_stats_collection: If True, also delete the statistics output collection.
                                  Defaults to False.

    Returns:
        bool: True if successful

    Example:
        >>> detach_statistics_function(collection, delete_stats_collection=True)
    """
    attached_fn = get_statistics_fn(collection)
    return collection.detach_function(
        attached_fn.name, delete_output_collection=delete_stats_collection
    )


def get_statistics(
    collection: "Collection",
    stats_collection_name: str,
    keys: Optional[OneOrMany[str]] = None,
) -> Dict[str, Any]:
    """Get the current statistics for a collection.

    Statistics include frequency counts for all metadata key-value pairs,
    as well as a summary with the total record count.

    Args:
        collection: The collection to get statistics for
        stats_collection_name: Name of the statistics collection to read from.
        keys: Optional metadata key(s) to filter statistics for. Can be a single key
              string or a list of keys. If provided, only returns statistics for
              those specific keys.

    Returns:
        Dict[str, Any]: A dictionary with the structure:
            {
                "statistics": {
                    "key1": {
                        "value1": {"count": count, ...},
                        "value2": {"count": count, ...}
                    },
                    "key2": {...},
                    ...
                },
                "summary": {
                    "total_count": count
                }
            }

    Example:
        >>> attach_statistics_function(collection, "my_collection_statistics")
        >>> collection.add(
        ...     ids=["id1", "id2"],
        ...     documents=["doc1", "doc2"],
        ...     metadatas=[{"category": "A", "score": 10}, {"category": "B", "score": 10}]
        ... )
        >>> # Wait for statistics to be computed
        >>> stats = get_statistics(collection, "my_collection_statistics")
        >>> print(stats)
        {
            "statistics": {
                "category": {
                    "A": {"count": 1},
                    "B": {"count": 1}
                },
                "score": {
                    "10": {"count": 2}
                }
            },
            "summary": {
                "total_count": 2
            }
        }

    Raises:
        ValueError: If more than 30 keys are provided in the keys filter.
    """
    # Normalize keys to list
    keys_list = maybe_cast_one_to_many(keys)

    # Validate keys count to avoid issues with large $in queries
    MAX_KEYS = 30
    if keys_list is not None and len(keys_list) > MAX_KEYS:
        raise ValueError(
            f"Too many keys provided: {len(keys_list)}. "
            f"Maximum allowed is {MAX_KEYS} keys per request. "
            "Consider calling get_statistics multiple times with smaller key batches."
        )

    # Import here to avoid circular dependency
    from chromadb.api.models.Collection import Collection

    # Get the statistics output collection model from the server
    stats_collection_model = collection._client.get_collection(
        name=stats_collection_name,
        tenant=collection.tenant,
        database=collection.database,
    )

    # Wrap it in a Collection object to access get/query methods
    stats_collection = Collection(
        client=collection._client,
        model=stats_collection_model,
        embedding_function=None,  # Statistics collections don't need embedding functions
        data_loader=None,
    )

    # Get all statistics records by paginating through the stats collection
    stats: Dict[str, Dict[str, Dict[str, int]]] = defaultdict(lambda: defaultdict(dict))
    summary: Dict[str, Any] = {}

    offset = 0
    # When filtering by keys, also include "summary" entries to get total_count
    where_filter: Optional[Where] = (
        cast(Where, {"key": {"$in": keys_list + ["summary"]}})
        if keys_list is not None
        else None
    )

    while True:
        page = stats_collection.get(
            include=["metadatas"], offset=offset, where=where_filter
        )

        metadatas = page.get("metadatas") or []
        if not metadatas:
            break

        for metadata in metadatas:
            if metadata is None:
                continue

            meta_key = metadata.get("key")
            value = metadata.get("value")
            value_label = metadata.get("value_label")
            value_type = metadata.get("type")
            count = metadata.get("count")

            if (
                meta_key is not None
                and value is not None
                and value_type is not None
                and count is not None
            ):
                if meta_key == "summary":
                    if value == "total_count":
                        summary["total_count"] = count
                else:
                    # Prioritize value_label if present, otherwise use value
                    stats_key = value_label if value_label is not None else value
                    assert isinstance(meta_key, str)
                    assert isinstance(stats_key, str)
                    assert isinstance(count, int)
                    stats[meta_key][stats_key]["count"] = count

        # Advance to next page using the actual number of items returned
        offset += len(metadatas)

    result = {"statistics": dict(stats)}
    if summary:
        result["summary"] = summary

    return result



"""
Integration test for Chroma's Task API

Tests the task creation, execution, and removal functionality
for automatically processing collections.
"""

import pytest
from chromadb.api.client import Client as ClientCreator
from chromadb.api.functions import (
    RECORD_COUNTER_FUNCTION,
    STATISTICS_FUNCTION,
    Function,
)
from chromadb.config import System
from chromadb.errors import ChromaError, NotFoundError
from chromadb.test.utils.wait_for_version_increase import (
    get_collection_version,
    wait_for_version_increase,
)
from time import sleep


def test_count_function_attach_and_detach(basic_http_client: System) -> None:
    """Test creating and removing a function with the record_counter operator"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    # Create a collection
    collection = client.get_or_create_collection(
        name="my_document",
        metadata={"description": "Sample documents for task processing"},
    )

    # Create a task that counts records in the collection
    attached_fn, created = collection.attach_function(
        name="count_my_docs",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="my_documents_counts",
        params=None,
    )

    # Verify task creation succeeded
    assert attached_fn is not None
    assert created is True
    initial_version = get_collection_version(client, collection.name)

    # Add documents
    collection.add(
        ids=["doc_{}".format(i) for i in range(0, 300)],
        documents=["test document"] * 300,
    )

    # Verify documents were added
    assert collection.count() == 300

    wait_for_version_increase(client, collection.name, initial_version)
    # Give some time to invalidate the frontend query cache
    sleep(60)

    result = client.get_collection("my_documents_counts").get("function_output")
    assert result["metadatas"] is not None
    assert result["metadatas"][0]["total_count"] == 300

    # Remove the task
    success = collection.detach_function(
        attached_fn.name,
        delete_output_collection=True,
    )

    # Verify task removal succeeded
    assert success is True


def test_task_with_invalid_function(basic_http_client: System) -> None:
    """Test that creating a task with an invalid function raises an error"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    collection = client.get_or_create_collection(name="test_invalid_function")
    collection.add(ids=["id1"], documents=["test document"])

    # Attempt to create task with non-existent function should raise ChromaError
    with pytest.raises(ChromaError, match="function not found"):
        collection.attach_function(
            function=Function._NONEXISTENT_TEST_ONLY,
            name="invalid_task",
            output_collection="output_collection",
            params=None,
        )


def test_attach_function_returns_function_name(basic_http_client: System) -> None:
    """Test that attach_function and get_attached_function return function_name field instead of UUID"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    collection = client.create_collection(name="test_function_name")
    collection.add(ids=["id1"], documents=["doc1"])

    # Attach a function and verify function_name field in response
    attached_fn, created = collection.attach_function(
        function=RECORD_COUNTER_FUNCTION,
        name="my_counter",
        output_collection="output_collection",
        params=None,
    )

    # Verify the attached function has function_name (not function_id UUID)
    assert created is True
    assert attached_fn.function_name == "record_counter"
    assert attached_fn.name == "my_counter"

    # Get the attached function and verify function_name field is also present
    retrieved_fn = collection.get_attached_function("my_counter")
    assert retrieved_fn == attached_fn

    # Clean up
    collection.detach_function(attached_fn.name, delete_output_collection=True)


def test_function_multiple_collections(basic_http_client: System) -> None:
    """Test attaching functions on multiple collections"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    # Create first collection and task
    collection1 = client.create_collection(name="collection_1")
    collection1.add(ids=["id1", "id2"], documents=["doc1", "doc2"])

    attached_fn1, created1 = collection1.attach_function(
        function=RECORD_COUNTER_FUNCTION,
        name="task_1",
        output_collection="output_1",
        params=None,
    )

    assert attached_fn1 is not None
    assert created1 is True

    # Create second collection and task
    collection2 = client.create_collection(name="collection_2")
    collection2.add(ids=["id3", "id4"], documents=["doc3", "doc4"])

    attached_fn2, created2 = collection2.attach_function(
        function=RECORD_COUNTER_FUNCTION,
        name="task_2",
        output_collection="output_2",
        params=None,
    )

    assert attached_fn2 is not None
    assert created2 is True

    # Task IDs should be different
    assert attached_fn1.id != attached_fn2.id

    # Clean up
    assert (
        collection1.detach_function(attached_fn1.name, delete_output_collection=True)
        is True
    )
    assert (
        collection2.detach_function(attached_fn2.name, delete_output_collection=True)
        is True
    )


def test_functions_one_attached_function_per_collection(
    basic_http_client: System,
) -> None:
    """Test that only one attached function is allowed per collection"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    # Create a single collection
    collection = client.create_collection(name="single_task_collection")
    collection.add(ids=["id1", "id2", "id3"], documents=["doc1", "doc2", "doc3"])

    # Create first task on the collection
    attached_fn1, created = collection.attach_function(
        function=RECORD_COUNTER_FUNCTION,
        name="task_1",
        output_collection="output_1",
        params=None,
    )

    assert attached_fn1 is not None
    assert created is True

    # Attempt to create a second task with a different name should fail
    # (only one attached function allowed per collection)
    with pytest.raises(
        ChromaError,
        match="collection already has an attached function: name=task_1, function=record_counter, output_collection=output_1",
    ):
        collection.attach_function(
            function=RECORD_COUNTER_FUNCTION,
            name="task_2",
            output_collection="output_2",
            params=None,
        )

    # Attempt to create a task with the same name but different function_id should also fail
    with pytest.raises(
        ChromaError,
        match=r"collection already has an attached function: name=task_1, function=record_counter, output_collection=output_1",
    ):
        collection.attach_function(
            function=STATISTICS_FUNCTION,
            name="task_1",
            output_collection="output_different",  # Different output collection
            params=None,
        )

    # Detach the first function
    assert (
        collection.detach_function(attached_fn1.name, delete_output_collection=True)
        is True
    )

    # Now we should be able to attach a new function
    attached_fn2, created2 = collection.attach_function(
        function=RECORD_COUNTER_FUNCTION,
        name="task_2",
        output_collection="output_2",
        params=None,
    )

    assert attached_fn2 is not None
    assert created2 is True
    assert attached_fn2.id != attached_fn1.id

    # Clean up
    assert (
        collection.detach_function(attached_fn2.name, delete_output_collection=True)
        is True
    )


def test_attach_function_with_invalid_params(basic_http_client: System) -> None:
    """Test that attach_function with non-empty params raises an error"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    collection = client.create_collection(name="test_invalid_params")
    collection.add(ids=["id1"], documents=["test document"])

    # Attempt to create task with non-empty params should fail
    # (no functions currently accept parameters)
    with pytest.raises(
        ChromaError,
        match="params must be empty - no functions currently accept parameters",
    ):
        collection.attach_function(
            name="invalid_params_task",
            function=RECORD_COUNTER_FUNCTION,
            output_collection="output_collection",
            params={"some_key": "some_value"},
        )


def test_attach_function_output_collection_already_exists(
    basic_http_client: System,
) -> None:
    """Test that attach_function fails when output collection name already exists"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    # Create a collection that will be used as input
    input_collection = client.create_collection(name="input_collection")
    input_collection.add(ids=["id1"], documents=["test document"])

    # Create another collection with the name we want to use for output
    client.create_collection(name="existing_output_collection")

    # Attempt to create task with output collection name that already exists
    with pytest.raises(
        ChromaError,
        match=r"Output collection \[existing_output_collection\] already exists",
    ):
        input_collection.attach_function(
            name="my_task",
            function=RECORD_COUNTER_FUNCTION,
            output_collection="existing_output_collection",
            params=None,
        )


def test_function_remove_nonexistent(basic_http_client: System) -> None:
    """Test removing a task that doesn't exist raises NotFoundError"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    collection = client.create_collection(name="test_collection")
    collection.add(ids=["id1"], documents=["test"])
    attached_fn, _ = collection.attach_function(
        function=RECORD_COUNTER_FUNCTION,
        name="test_function",
        output_collection="output_collection",
        params=None,
    )

    collection.detach_function(attached_fn.name, delete_output_collection=True)

    # Trying to detach this function again should raise NotFoundError
    with pytest.raises(NotFoundError, match="does not exist"):
        collection.detach_function(attached_fn.name, delete_output_collection=True)


def test_attach_to_output_collection_fails(basic_http_client: System) -> None:
    """Test that attaching a function to an output collection fails"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    # Create input collection
    input_collection = client.create_collection(name="input_collection")
    input_collection.add(ids=["id1"], documents=["test"])

    _, _ = input_collection.attach_function(
        name="test_function",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="output_collection",
        params=None,
    )
    output_collection = client.get_collection(name="output_collection")

    with pytest.raises(
        ChromaError, match="cannot attach function to an output collection"
    ):
        _ = output_collection.attach_function(
            name="test_function_2",
            function=RECORD_COUNTER_FUNCTION,
            output_collection="output_collection_2",
            params=None,
        )


def test_delete_output_collection_detaches_function(basic_http_client: System) -> None:
    """Test that deleting an output collection also detaches the attached function"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    # Create input collection and attach a function
    input_collection = client.create_collection(name="input_collection")
    input_collection.add(ids=["id1"], documents=["test"])

    attached_fn, created = input_collection.attach_function(
        name="my_function",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="output_collection",
        params=None,
    )
    assert attached_fn is not None
    assert created is True

    # Delete the output collection directly
    client.delete_collection("output_collection")

    # The attached function should now be gone - trying to get it should raise NotFoundError
    with pytest.raises(NotFoundError):
        input_collection.get_attached_function("my_function")


def test_delete_orphaned_output_collection(basic_http_client: System) -> None:
    """Test that deleting an output collection from a recently detached function works"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    # Create input collection and attach a function
    input_collection = client.create_collection(name="input_collection")
    input_collection.add(ids=["id1"], documents=["test"])

    attached_fn, created = input_collection.attach_function(
        name="my_function",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="output_collection",
        params=None,
    )
    assert attached_fn is not None
    assert created is True

    input_collection.detach_function(attached_fn.name, delete_output_collection=False)

    # Delete the output collection directly
    client.delete_collection("output_collection")

    # The attached function should still exist but be marked as detached
    with pytest.raises(NotFoundError):
        input_collection.get_attached_function("my_function")

    with pytest.raises(NotFoundError):
        # Try to use the function - it should fail since it's detached
        client.get_collection("output_collection")


def test_partial_attach_function_repair(
    basic_http_client: System,
) -> None:
    """Test creating and removing a function with the record_counter operator"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    # Create a collection
    collection = client.get_or_create_collection(
        name="my_document",
    )

    # Create a task that counts records in the collection
    attached_fn, created = collection.attach_function(
        name="count_my_docs",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="my_documents_counts",
        params=None,
    )
    assert created is True

    # Verify task creation succeeded
    assert attached_fn is not None

    collection2 = client.get_or_create_collection(
        name="my_document2",
    )

    # Create a task that counts records in the collection
    # This should fail
    with pytest.raises(
        ChromaError, match=r"Output collection \[my_documents_counts\] already exists"
    ):
        attached_fn, _ = collection2.attach_function(
            name="count_my_docs",
            function=RECORD_COUNTER_FUNCTION,
            output_collection="my_documents_counts",
            params=None,
        )

    # Detach the function
    assert (
        collection.detach_function(attached_fn.name, delete_output_collection=True)
        is True
    )

    # Create a task that counts records in the collection
    attached_fn, created = collection2.attach_function(
        name="count_my_docs",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="my_documents_counts",
        params=None,
    )
    assert attached_fn is not None
    assert created is True


def test_output_collection_created_with_schema(basic_http_client: System) -> None:
    """Test that output collections are created with the source_attached_function_id in the schema"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    # Create input collection and attach a function
    input_collection = client.create_collection(name="input_collection")
    input_collection.add(ids=["id1"], documents=["test"])

    attached_fn, created = input_collection.attach_function(
        name="my_function",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="output_collection",
        params=None,
    )
    assert attached_fn is not None
    assert created is True

    # Get the output collection - it should exist
    output_collection = client.get_collection(name="output_collection")
    assert output_collection is not None

    # The source_attached_function_id is stored in the schema (not metadata)
    # We can't directly access the schema from the client, but we verify the collection exists
    # and the attached function orchestrator will use this field internally
    assert "source_attached_function_id" in output_collection._model.pretty_schema()

    # Clean up
    input_collection.detach_function(attached_fn.name, delete_output_collection=True)


def test_count_function_attach_and_detach_attach_attach(
    basic_http_client: System,
) -> None:
    """Test creating and removing a function with the record_counter operator"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    # Create a collection
    collection = client.get_or_create_collection(
        name="my_document",
        metadata={"description": "Sample documents for task processing"},
    )

    # Create a task that counts records in the collection
    attached_fn, created = collection.attach_function(
        name="count_my_docs",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="my_documents_counts",
        params=None,
    )

    # Verify task creation succeeded
    assert created is True
    assert attached_fn is not None
    initial_version = get_collection_version(client, collection.name)

    # Add documents
    collection.add(
        ids=["doc_{}".format(i) for i in range(0, 300)],
        documents=["test document"] * 300,
    )

    # Verify documents were added
    assert collection.count() == 300

    wait_for_version_increase(client, collection.name, initial_version)
    # Give some time to invalidate the frontend query cache
    sleep(60)

    result = client.get_collection("my_documents_counts").get("function_output")
    assert result["metadatas"] is not None
    assert result["metadatas"][0]["total_count"] == 300

    # Remove the task
    success = collection.detach_function(
        attached_fn.name, delete_output_collection=True
    )

    # Verify task removal succeeded
    assert success is True

    # Attach a function that counts records in the collection
    attached_fn, created = collection.attach_function(
        name="count_my_docs",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="my_documents_counts",
        params=None,
    )
    assert attached_fn is not None
    assert created is True

    # Attach a function that counts records in the collection
    attached_fn, created = collection.attach_function(
        name="count_my_docs",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="my_documents_counts",
        params=None,
    )
    assert created is False
    assert attached_fn is not None


def test_attach_function_idempotency(basic_http_client: System) -> None:
    """Test that attach_function is idempotent - calling it twice with same params returns created=False"""
    client = ClientCreator.from_system(basic_http_client)
    client.reset()

    collection = client.create_collection(name="idempotency_test")
    collection.add(ids=["id1"], documents=["test document"])

    # First attach - should be newly created
    attached_fn1, created1 = collection.attach_function(
        name="my_function",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="output_collection",
        params=None,
    )
    assert attached_fn1 is not None
    assert created1 is True

    # Second attach with identical params - should be idempotent (created=False)
    attached_fn2, created2 = collection.attach_function(
        name="my_function",
        function=RECORD_COUNTER_FUNCTION,
        output_collection="output_collection",
        params=None,
    )
    assert attached_fn2 is not None
    assert created2 is False

    # Both should return the same function ID
    assert attached_fn1.id == attached_fn2.id

    # Clean up
    collection.detach_function(attached_fn1.name, delete_output_collection=True)



from chromadb.config import Component, System, Settings
from overrides import overrides
from threading import local
import random

data = local()  # use thread local just in case tests ever run in parallel


def reset() -> None:
    global data
    data.starts = []
    data.stops = []
    data.inits = []


class ComponentA(Component):
    def __init__(self, system: System):
        data.inits += "A"
        super().__init__(system)
        self.require(ComponentB)
        self.require(ComponentC)

    @overrides
    def start(self) -> None:
        data.starts += "A"

    @overrides
    def stop(self) -> None:
        data.stops += "A"


class ComponentB(Component):
    def __init__(self, system: System):
        data.inits += "B"
        super().__init__(system)
        self.require(ComponentC)
        self.require(ComponentD)

    @overrides
    def start(self) -> None:
        data.starts += "B"

    @overrides
    def stop(self) -> None:
        data.stops += "B"


class ComponentC(Component):
    def __init__(self, system: System):
        data.inits += "C"
        super().__init__(system)
        self.require(ComponentD)

    @overrides
    def start(self) -> None:
        data.starts += "C"

    @overrides
    def stop(self) -> None:
        data.stops += "C"


class ComponentD(Component):
    def __init__(self, system: System):
        data.inits += "D"
        super().__init__(system)

    @overrides
    def start(self) -> None:
        data.starts += "D"

    @overrides
    def stop(self) -> None:
        data.stops += "D"


# Dependency Graph for tests:
# ┌───┐
# │ A │
# └┬─┬┘
#  │┌▽──┐
#  ││ B │
#  │└┬─┬┘
# ┌▽─▽┐│
# │ C ││
# └┬──┘│
# ┌▽───▽┐
# │  D  │
# └─────┘


def test_leaf_only() -> None:
    settings = Settings()
    system = System(settings)

    reset()

    d = system.instance(ComponentD)
    assert isinstance(d, ComponentD)

    assert data.inits == ["D"]
    system.start()
    assert data.starts == ["D"]
    system.stop()
    assert data.stops == ["D"]


def test_partial() -> None:
    settings = Settings()
    system = System(settings)

    reset()

    c = system.instance(ComponentC)
    assert isinstance(c, ComponentC)

    assert data.inits == ["C", "D"]
    system.start()
    assert data.starts == ["D", "C"]
    system.stop()
    assert data.stops == ["C", "D"]


def test_system_startup() -> None:
    settings = Settings()
    system = System(settings)

    reset()

    a = system.instance(ComponentA)
    assert isinstance(a, ComponentA)

    assert data.inits == ["A", "B", "C", "D"]
    system.start()
    assert data.starts == ["D", "C", "B", "A"]
    system.stop()
    assert data.stops == ["A", "B", "C", "D"]


def test_system_override_order() -> None:
    settings = Settings()
    system = System(settings)

    reset()

    system.instance(ComponentA)

    # Deterministically shuffle the instances map to prove that topsort is actually
    # working and not just implicitly working because of insertion order.

    # This causes the test to actually fail if the deps are not wired up correctly.
    random.seed(0)
    entries = list(system._instances.items())
    random.shuffle(entries)
    system._instances = {k: v for k, v in entries}

    system.start()
    assert data.starts == ["D", "C", "B", "A"]
    system.stop()
    assert data.stops == ["A", "B", "C", "D"]


class ComponentZ(Component):
    def __init__(self, system: System):
        super().__init__(system)
        self.require(ComponentC)

    @overrides
    def start(self) -> None:
        pass

    @overrides
    def stop(self) -> None:
        pass


def test_runtime_dependencies() -> None:
    settings = Settings()
    system = System(settings)

    reset()

    # Nothing to do, no components were requested prior to start
    system.start()
    assert data.starts == []

    # Constructs dependencies and starts them in the correct order
    ComponentZ(system)
    assert data.starts == ["D", "C"]
    system.stop()
    assert data.stops == ["C", "D"]


def test_http_client_setting_defaults() -> None:
    settings = Settings()
    assert settings.chroma_http_keepalive_secs == 40.0
    assert settings.chroma_http_max_connections is None
    assert settings.chroma_http_max_keepalive_connections is None


def test_http_client_setting_overrides() -> None:
    settings = Settings(
        chroma_http_keepalive_secs=5.5,
        chroma_http_max_connections=123,
        chroma_http_max_keepalive_connections=17,
    )
    assert settings.chroma_http_keepalive_secs == 5.5
    assert settings.chroma_http_max_connections == 123
    assert settings.chroma_http_max_keepalive_connections == 17



from typing import Any, Dict, Mapping, Optional, TypeVar
from urllib.parse import quote, urlparse, urlunparse
import logging
import orjson as json
import httpx

import chromadb.errors as errors
from chromadb.config import Component, Settings, System

logger = logging.getLogger(__name__)


# inherits from Component so that it can create an init function to use system
# this way it can build limits from the settings in System
class BaseHTTPClient(Component):
    _settings: Settings
    pre_flight_checks: Any = None
    DEFAULT_KEEPALIVE_SECS: float = 40.0

    def __init__(self, system: System):
        super().__init__(system)
        self._settings = system.settings
        keepalive_setting = self._settings.chroma_http_keepalive_secs
        self.keepalive_secs: Optional[float] = (
            keepalive_setting
            if keepalive_setting is not None
            else BaseHTTPClient.DEFAULT_KEEPALIVE_SECS
        )
        self._http_limits = self._build_limits()

    def _build_limits(self) -> httpx.Limits:
        limit_kwargs: Dict[str, Any] = {}
        if self.keepalive_secs is not None:
            limit_kwargs["keepalive_expiry"] = self.keepalive_secs

        max_connections = self._settings.chroma_http_max_connections
        if max_connections is not None:
            limit_kwargs["max_connections"] = max_connections

        max_keepalive_connections = self._settings.chroma_http_max_keepalive_connections
        if max_keepalive_connections is not None:
            limit_kwargs["max_keepalive_connections"] = max_keepalive_connections

        return httpx.Limits(**limit_kwargs)

    @property
    def http_limits(self) -> httpx.Limits:
        return self._http_limits

    @staticmethod
    def _validate_host(host: str) -> None:
        parsed = urlparse(host)
        if "/" in host and parsed.scheme not in {"http", "https"}:
            raise ValueError(
                "Invalid URL. " f"Unrecognized protocol - {parsed.scheme}."
            )
        if "/" in host and (not host.startswith("http")):
            raise ValueError(
                "Invalid URL. "
                "Seems that you are trying to pass URL as a host but without \
                  specifying the protocol. "
                "Please add http:// or https:// to the host."
            )

    @staticmethod
    def resolve_url(
        chroma_server_host: str,
        chroma_server_ssl_enabled: Optional[bool] = False,
        default_api_path: Optional[str] = "",
        chroma_server_http_port: Optional[int] = 8000,
    ) -> str:
        _skip_port = False
        _chroma_server_host = chroma_server_host
        BaseHTTPClient._validate_host(_chroma_server_host)
        if _chroma_server_host.startswith("http"):
            logger.debug("Skipping port as the user is passing a full URL")
            _skip_port = True
        parsed = urlparse(_chroma_server_host)

        scheme = "https" if chroma_server_ssl_enabled else parsed.scheme or "http"
        net_loc = parsed.netloc or parsed.hostname or chroma_server_host
        port = (
            ":" + str(parsed.port or chroma_server_http_port) if not _skip_port else ""
        )
        path = parsed.path or default_api_path

        if not path or path == net_loc:
            path = default_api_path if default_api_path else ""
        if not path.endswith(default_api_path or ""):
            path = path + default_api_path if default_api_path else ""
        full_url = urlunparse(
            (scheme, f"{net_loc}{port}", quote(path.replace("//", "/")), "", "", "")
        )

        return full_url

    # requests removes None values from the built query string, but httpx includes it as an empty value
    T = TypeVar("T", bound=Dict[Any, Any])

    @staticmethod
    def _clean_params(params: T) -> T:
        """Remove None values from provided dict."""
        return {k: v for k, v in params.items() if v is not None}  # type: ignore

    @staticmethod
    def _raise_chroma_error(resp: httpx.Response) -> None:
        """Raises an error if the response is not ok, using a ChromaError if possible."""
        try:
            resp.raise_for_status()
            return
        except httpx.HTTPStatusError:
            pass

        chroma_error = None
        try:
            body = json.loads(resp.text)
            if "error" in body:
                if body["error"] in errors.error_types:
                    chroma_error = errors.error_types[body["error"]](body["message"])

                    trace_id = resp.headers.get("chroma-trace-id")
                    if trace_id:
                        chroma_error.trace_id = trace_id

        except BaseException:
            pass

        if chroma_error:
            raise chroma_error

        try:
            resp.raise_for_status()
        except httpx.HTTPStatusError:
            trace_id = resp.headers.get("chroma-trace-id")
            if trace_id:
                raise Exception(f"{resp.text} (trace ID: {trace_id})")
            raise (Exception(resp.text))

    def get_request_headers(self) -> Mapping[str, str]:
        """Return headers used for HTTP requests."""
        return {}

    def get_api_url(self) -> str:
        """Return the API URL for this client."""
        return ""



import re
import sys

import chromadb_rust_bindings
import requests
from packaging.version import parse

import chromadb


def build_cli_args(**kwargs):
    args = []
    for key, value in kwargs.items():
        if isinstance(value, bool):
            if value:
                args.append(f"--{key}")
        elif value is not None:
            args.extend([f"--{key}", str(value)])
    return args


def update():
    try:
        url = f"https://api.github.com/repos/chroma-core/chroma/releases"
        response = requests.get(url)
        response.raise_for_status()
        releases = response.json()

        version_pattern = re.compile(r'^\d+\.\d+\.\d+$')
        numeric_releases = [r["tag_name"] for r in releases if version_pattern.fullmatch(r["tag_name"])]

        if not numeric_releases:
            print("Couldn't fetch the latest Chroma version")
            return

        latest = max(numeric_releases, key=parse)
        if latest == chromadb.__version__:
            print("Your Chroma version is up-to-date")
            return

        print(
            f"A new version of Chroma is available!\nIf you're using pip, run 'pip install --upgrade chromadb' to upgrade to version {latest}")

    except Exception as e:
        print("Couldn't fetch the latest Chroma version")


def app():
    args = sys.argv
    if ["chroma", "update"] in args:
        update()
        return
    try:
        chromadb_rust_bindings.cli(args)
    except KeyboardInterrupt:
        pass


import math
from concurrent.futures import ThreadPoolExecutor, as_completed

import pytest

from chromadb import SparseVector
from chromadb.utils.embedding_functions.chroma_bm25_embedding_function import (
    DEFAULT_CHROMA_BM25_STOPWORDS,
    ChromaBm25EmbeddingFunction,
)


def _is_sorted(values: list[int]) -> bool:
    return all(values[i] >= values[i - 1] for i in range(1, len(values)))


def test_comprehensive_tokenization_matches_reference() -> None:
    embedder = ChromaBm25EmbeddingFunction()
    embedding = embedder(
        [
            "Usain Bolt's top speed reached ~27.8 mph (44.72 km/h)",
        ]
    )[0]

    expected_indices = [
        230246813,
        395514983,
        458027949,
        488165615,
        729632045,
        734978415,
        997512866,
        1114505193,
        1381820790,
        1501587190,
        1649421877,
        1837285388,
    ]
    expected_value = 1.6391153

    assert embedding.indices == expected_indices
    for value in embedding.values:
        assert value == pytest.approx(expected_value, abs=1e-5)


def test_matches_rust_reference_values() -> None:
    embedder = ChromaBm25EmbeddingFunction()
    embedding = embedder(
        [
            "The   space-time   continuum   WARPS   near   massive   objects...",
        ]
    )[0]

    expected_indices = [
        90097469,
        519064992,
        737893654,
        1110755108,
        1950894484,
        2031641008,
        2058513491,
    ]
    expected_value = 1.660867

    assert embedding.indices == expected_indices
    for value in embedding.values:
        assert value == pytest.approx(expected_value, abs=1e-5)


def test_generates_embeddings_for_multiple_documents() -> None:
    embedder = ChromaBm25EmbeddingFunction()
    texts = [
        "Usain Bolt's top speed reached ~27.8 mph (44.72 km/h)",
        "The   space-time   continuum   WARPS   near   massive   objects...",
        "BM25 is great for sparse retrieval tasks",
    ]

    embeddings = embedder(texts)

    assert len(embeddings) == len(texts)
    for embedding in embeddings:
        assert embedding.indices
        assert len(embedding.indices) == len(embedding.values)
        assert _is_sorted(embedding.indices)
        for value in embedding.values:
            assert value > 0
            assert math.isfinite(value)


def test_embed_query_matches_call() -> None:
    embedder = ChromaBm25EmbeddingFunction()
    query = "retrieve BM25 docs"

    query_embedding = embedder.embed_query([query])[0]
    doc_embedding = embedder([query])[0]

    assert query_embedding.indices == doc_embedding.indices
    assert query_embedding.values == doc_embedding.values


def test_config_round_trip() -> None:
    embedder = ChromaBm25EmbeddingFunction()
    config = embedder.get_config()

    assert config["k"] == pytest.approx(1.2, abs=1e-9)
    assert config["b"] == pytest.approx(0.75, abs=1e-9)
    assert config["avg_doc_length"] == pytest.approx(256.0, abs=1e-9)
    assert config["token_max_length"] == 40
    assert "stopwords" not in config

    custom_stopwords = DEFAULT_CHROMA_BM25_STOPWORDS[:10]
    rebuilt = ChromaBm25EmbeddingFunction.build_from_config(
        {
            **config,
            "stopwords": custom_stopwords,
        }
    )

    rebuilt_config = rebuilt.get_config()
    assert rebuilt_config["stopwords"] == custom_stopwords
    assert rebuilt_config["token_max_length"] == config["token_max_length"]
    assert rebuilt_config["k"] == pytest.approx(config["k"], abs=1e-9)
    assert rebuilt_config["b"] == pytest.approx(config["b"], abs=1e-9)
    assert rebuilt_config["avg_doc_length"] == pytest.approx(
        config["avg_doc_length"], abs=1e-9
    )


def test_validate_config_update_rejects_unknown_keys() -> None:
    embedder = ChromaBm25EmbeddingFunction()

    with pytest.raises(ValueError):
        embedder.validate_config_update(embedder.get_config(), {"unknown": 123})


def test_validate_config_update_allows_known_keys() -> None:
    embedder = ChromaBm25EmbeddingFunction()

    embedder.validate_config_update(
        embedder.get_config(), {"k": 1.1, "stopwords": ["custom"]}
    )


def test_multithreaded_usage() -> None:
    embedder = ChromaBm25EmbeddingFunction()
    base_texts = [
        """The gravitational wave background from massive black hole binaries emit bursts of
        gravitational waves at periapse. Such events may be directly resolvable in the Galactic
        centre. However, if the star does not spiral in, the emitted GWs are not resolvable for
        extra-galactic MBHs, but constitute a source of background noise. We estimate the power
        spectrum of this extreme mass ratio burst background.""",
        """Dynamics of planets in exoplanetary systems with multiple stars showing how the
        gravitational interactions between the stars and planets affect the orbital stability
        and long-term evolution of the planetary system architectures.""",
        """Diurnal Thermal Tides in a Non-rotating atmosphere with realistic heating profiles
        and temperature gradients that demonstrate the complex interplay between radiation
        and atmospheric dynamics in planetary atmospheres.""",
        """Intermittent turbulence, noise and waves in stellar atmospheres create complex
        patterns of energy transport and momentum deposition that influence the structure
        and evolution of stellar interiors and surfaces.""",
        """Superconductivity in quantum materials and condensed matter physics systems
        exhibiting novel quantum phenomena including topological phases, strongly correlated
        electron systems, and exotic superconducting pairing mechanisms.""",
        """Machine learning models require careful tuning of hyperparameters including learning
        rates, regularization coefficients, and architectural choices that demonstrate the
        complex interplay between optimization algorithms and model capacity.""",
        """Natural language processing enables text understanding through sophisticated
        algorithms that analyze semantic relationships, syntactic structures, and contextual
        information to extract meaningful representations from unstructured textual data.""",
        """Vector databases store high-dimensional embeddings efficiently using advanced
        indexing techniques including approximate nearest neighbor search algorithms that
        balance accuracy and computational efficiency for large-scale similarity search.""",
    ]
    texts = base_texts * 30

    num_threads = 10

    def process_single_text(text: str) -> SparseVector:
        return embedder([text])[0]

    with ThreadPoolExecutor(max_workers=num_threads) as executor:
        futures = [executor.submit(process_single_text, text) for text in texts]
        all_results = []
        for future in as_completed(futures):
            try:
                embedding = future.result()
                all_results.append(embedding)
            except Exception as e:
                pytest.fail(
                    f"Threading error detected: {type(e).__name__}: {e}. "
                    "This indicates the stemmer is not thread-safe when cached."
                )

    assert len(all_results) == len(texts)

    for embedding in all_results:
        assert embedding.indices
        assert len(embedding.indices) == len(embedding.values)
        assert _is_sorted(embedding.indices)
        for value in embedding.values:
            assert value > 0
            assert math.isfinite(value)



from chromadb.api.types import Embeddings, Documents, EmbeddingFunction, Space
from chromadb.utils.embedding_functions.schemas import validate_config_schema
from typing import List, Dict, Any
import os
import numpy as np


class MistralEmbeddingFunction(EmbeddingFunction[Documents]):
    def __init__(
        self,
        model: str,
        api_key_env_var: str = "MISTRAL_API_KEY",
    ):
        """
        Initialize the MistralEmbeddingFunction.

        Args:
            model (str): The name of the model to use for text embeddings.
            api_key_env_var (str): The environment variable name for the Mistral API key.
        """
        try:
            from mistralai import Mistral
        except ImportError:
            raise ValueError(
                "The mistralai python package is not installed. Please install it with `pip install mistralai`"
            )
        self.model = model
        self.api_key_env_var = api_key_env_var
        self.api_key = os.getenv(api_key_env_var)
        if not self.api_key:
            raise ValueError(f"The {api_key_env_var} environment variable is not set.")
        self.client = Mistral(api_key=self.api_key)

    def __call__(self, input: Documents) -> Embeddings:
        """
        Get the embeddings for a list of texts.

        Args:
            input (Documents): A list of texts to get embeddings for.
        """
        if not all(isinstance(item, str) for item in input):
            raise ValueError("Mistral only supports text documents, not images")
        output = self.client.embeddings.create(
            model=self.model,
            inputs=input,
        )

        # Extract embeddings from the response
        return [np.array(data.embedding) for data in output.data]

    @staticmethod
    def name() -> str:
        return "mistral"

    def default_space(self) -> Space:
        return "cosine"

    def supported_spaces(self) -> List[Space]:
        return ["cosine", "l2", "ip"]

    @staticmethod
    def build_from_config(config: Dict[str, Any]) -> "EmbeddingFunction[Documents]":
        model = config.get("model")
        api_key_env_var = config.get("api_key_env_var")

        if model is None or api_key_env_var is None:
            assert False, "This code should not be reached"  # this is for type checking
        return MistralEmbeddingFunction(model=model, api_key_env_var=api_key_env_var)

    def get_config(self) -> Dict[str, Any]:
        return {
            "model": self.model,
            "api_key_env_var": self.api_key_env_var,
        }

    def validate_config_update(
        self, old_config: Dict[str, Any], new_config: Dict[str, Any]
    ) -> None:
        if "model" in new_config:
            raise ValueError(
                "The model cannot be changed after the embedding function has been initialized."
            )

    @staticmethod
    def validate_config(config: Dict[str, Any]) -> None:
        """
        Validate the configuration using the JSON schema.

        Args:
            config: Configuration to validate
        """
        validate_config_schema(config, "mistral")



from chromadb.utils.embedding_functions.schemas.schema_utils import (
    validate_config_schema,
    load_schema,
    get_schema_version,
)
from chromadb.utils.embedding_functions.schemas.registry import (
    get_available_schemas,
    get_schema_info,
    get_embedding_function_names,
)

__all__ = [
    "validate_config_schema",
    "load_schema",
    "get_schema_version",
    "get_available_schemas",
    "get_schema_info",
    "get_embedding_function_names",
]



import os
import random
import gc
import time


# Borrowed from https://github.com/rogerbinns/apsw/blob/master/apsw/tests.py#L224
# Used to delete sqlite files on Windows, since Windows file locking
# behaves differently to other operating systems
# This should only be used for test or non-production code, such as in reset_state.
def delete_file(name: str) -> None:
    try:
        os.remove(name)
    except Exception:
        pass

    chars = list("abcdefghijklmn")
    random.shuffle(chars)
    newname = name + "-n-" + "".join(chars)
    count = 0
    while os.path.exists(name):
        count += 1
        try:
            os.rename(name, newname)
        except Exception:
            if count > 30:
                n = list("abcdefghijklmnopqrstuvwxyz")
                random.shuffle(n)
                final_name = "".join(n)
                try:
                    os.rename(
                        name, "chroma-to-clean" + final_name + ".deletememanually"
                    )
                except Exception:
                    pass
                break
            time.sleep(0.1)
            gc.collect()



from chromadb.api.types import (
    SparseEmbeddingFunction,
    SparseVectors,
    Documents,
)
from typing import Dict, Any, TypedDict, Optional
from typing import cast, Literal
import warnings
from chromadb.utils.embedding_functions.schemas import validate_config_schema
from chromadb.utils.sparse_embedding_utils import normalize_sparse_vector

TaskType = Literal["document", "query"]


class Bm25EmbeddingFunctionQueryConfig(TypedDict):
    task: TaskType


class Bm25EmbeddingFunction(SparseEmbeddingFunction[Documents]):
    def __init__(
        self,
        avg_len: Optional[float] = None,
        task: Optional[TaskType] = "document",
        cache_dir: Optional[str] = None,
        k: Optional[float] = None,
        b: Optional[float] = None,
        language: Optional[str] = None,
        token_max_length: Optional[int] = None,
        disable_stemmer: Optional[bool] = None,
        specific_model_path: Optional[str] = None,
        query_config: Optional[Bm25EmbeddingFunctionQueryConfig] = None,
        **kwargs: Any,
    ):
        """Initialize SparseEncoderEmbeddingFunction.

        Args:
            avg_len(float, optional): The average length of the documents in the corpus.
            task (str, optional): Task to perform, can be "document" or "query"
            cache_dir (str, optional): The path to the cache directory.
            k (float, optional): The k parameter in the BM25 formula. Defines the saturation of the term frequency.
            b (float, optional): The b parameter in the BM25 formula. Defines the importance of the document length.
            language (str, optional): Specifies the language for the stemmer.
            token_max_length (int, optional): The maximum length of the tokens.
            disable_stemmer (bool, optional): Disable the stemmer.
            specific_model_path (str, optional): The path to the specific model.
            query_config (dict, optional): Configuration for the query, can be "task"
            **kwargs: Additional arguments to pass to the Bm25 model.
        """
        warnings.warn(
            "Bm25EmbeddingFunction is deprecated. Please use ChromaBm25EmbeddingFunction instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        try:
            from fastembed.sparse.bm25 import Bm25
        except ImportError:
            raise ValueError(
                "The fastembed python package is not installed. Please install it with `pip install fastembed`"
            )

        self.task = task
        self.query_config = query_config
        self.cache_dir = cache_dir
        self.k = k
        self.b = b
        self.avg_len = avg_len
        self.language = language
        self.token_max_length = token_max_length
        self.disable_stemmer = disable_stemmer
        self.specific_model_path = specific_model_path
        for key, value in kwargs.items():
            if not isinstance(value, (str, int, float, bool, list, dict, tuple)):
                raise ValueError(f"Keyword argument {key} is not a primitive type")
        self.kwargs = kwargs
        bm25_kwargs = {
            "model_name": "Qdrant/bm25",
        }
        optional_params = {
            "cache_dir": cache_dir,
            "k": k,
            "b": b,
            "avg_len": avg_len,
            "language": language,
            "token_max_length": token_max_length,
            "disable_stemmer": disable_stemmer,
            "specific_model_path": specific_model_path,
        }
        for key, value in optional_params.items():
            if value is not None:
                bm25_kwargs[key] = value
        bm25_kwargs.update({k: v for k, v in kwargs.items() if v is not None})
        self._model = Bm25(**bm25_kwargs)

    def __call__(self, input: Documents) -> SparseVectors:
        """Generate embeddings for the given documents.

        Args:
            input: Documents to generate embeddings for.

        Returns:
            Embeddings for the documents.
        """
        try:
            from fastembed.sparse.bm25 import Bm25
        except ImportError:
            raise ValueError(
                "The fastembed python package is not installed. Please install it with `pip install fastembed`"
            )
        model = cast(Bm25, self._model)
        if self.task == "document":
            embeddings = model.embed(
                list(input),
            )
        elif self.task == "query":
            embeddings = model.query_embed(
                list(input),
            )
        else:
            raise ValueError(f"Invalid task: {self.task}")

        sparse_vectors: SparseVectors = []

        for vec in embeddings:
            sparse_vectors.append(
                normalize_sparse_vector(
                    indices=vec.indices.tolist(), values=vec.values.tolist()
                )
            )

        return sparse_vectors

    def embed_query(self, input: Documents) -> SparseVectors:
        try:
            from fastembed.sparse.bm25 import Bm25
        except ImportError:
            raise ValueError(
                "The fastembed python package is not installed. Please install it with `pip install fastembed`"
            )
        model = cast(Bm25, self._model)
        if self.query_config is not None:
            task = self.query_config.get("task")
            if task == "document":
                embeddings = model.embed(
                    list(input),
                )
            elif task == "query":
                embeddings = model.query_embed(
                    list(input),
                )
            else:
                raise ValueError(f"Invalid task: {task}")

            sparse_vectors: SparseVectors = []

            for vec in embeddings:
                sparse_vectors.append(
                    normalize_sparse_vector(
                        indices=vec.indices.tolist(), values=vec.values.tolist()
                    )
                )

            return sparse_vectors

        else:
            return self.__call__(input)

    @staticmethod
    def name() -> str:
        return "bm25"

    @staticmethod
    def build_from_config(
        config: Dict[str, Any]
    ) -> "SparseEmbeddingFunction[Documents]":
        task = config.get("task")
        query_config = config.get("query_config")
        cache_dir = config.get("cache_dir")
        k = config.get("k")
        b = config.get("b")
        avg_len = config.get("avg_len")
        language = config.get("language")
        token_max_length = config.get("token_max_length")
        disable_stemmer = config.get("disable_stemmer")
        specific_model_path = config.get("specific_model_path")
        kwargs = config.get("kwargs", {})

        return Bm25EmbeddingFunction(
            task=task,
            query_config=query_config,
            cache_dir=cache_dir,
            k=k,
            b=b,
            avg_len=avg_len,
            language=language,
            token_max_length=token_max_length,
            disable_stemmer=disable_stemmer,
            specific_model_path=specific_model_path,
            **kwargs,
        )

    def get_config(self) -> Dict[str, Any]:
        return {
            "task": self.task,
            "query_config": self.query_config,
            "cache_dir": self.cache_dir,
            "k": self.k,
            "b": self.b,
            "avg_len": self.avg_len,
            "language": self.language,
            "token_max_length": self.token_max_length,
            "disable_stemmer": self.disable_stemmer,
            "specific_model_path": self.specific_model_path,
            "kwargs": self.kwargs,
        }

    def validate_config_update(
        self, old_config: Dict[str, Any], new_config: Dict[str, Any]
    ) -> None:
        # Users should be able to change the path if needed, so we should not validate that.
        # e.g. moving file path from /v1/my-model.bin to /v2/my-model.bin
        return

    @staticmethod
    def validate_config(config: Dict[str, Any]) -> None:
        """
        Validate the configuration using the JSON schema.

        Args:
            config: Configuration to validate

        Raises:
            ValidationError: If the configuration does not match the schema
        """
        validate_config_schema(config, "bm25")


from typing import Dict, List
from hypothesis import given
from chromadb.test.conftest import (
    ClientFactories,
)
import hypothesis.strategies as st
from chromadb.test.conftest import MULTI_REGION_ENABLED


def test_list_databases(client_factories: ClientFactories) -> None:
    client = client_factories.create_client()
    client.reset()
    admin_client = client_factories.create_admin_client_from_system()

    for i in range(10):
        admin_client.create_database(f"test_list_databases_{i}")

    databases = admin_client.list_databases()
    total_default_databases = 2 if MULTI_REGION_ENABLED else 1
    assert len(databases) == 10 + total_default_databases

    for i in range(10):
        assert any(d["name"] == f"test_list_databases_{i}" for d in databases)

    assert any(d["name"] == "default_database" for d in databases)

    if MULTI_REGION_ENABLED:
        assert any(d["name"] == "tilt-spanning+default_database" for d in databases)


@st.composite
def tenants_and_databases_st(
    draw: st.DrawFn, max_tenants: int, max_databases: int
) -> Dict[str, List[str]]:
    """Generates a set of random tenants and databases. Each database is assigned to a random tenant. Returns a dictionary where the key is the tenant name and the value is a list of database names for that tenant."""
    num_tenants = draw(st.integers(min_value=1, max_value=max_tenants))
    num_databases = draw(st.integers(min_value=0, max_value=max_databases))

    database_i_to_tenant_i = draw(
        st.lists(
            st.integers(min_value=0, max_value=num_tenants - 1),
            min_size=num_databases,
            max_size=num_databases,
        )
    )

    tenants = [f"tenant_{i}" for i in range(num_tenants)]
    databases = [f"database_{i}" for i in range(num_databases)]

    result: Dict[str, List[str]] = {}
    for database_i, tenant_i in enumerate(database_i_to_tenant_i):
        tenant = tenants[tenant_i]
        database = databases[database_i]

        if tenant not in result:
            result[tenant] = []

        result[tenant].append(database)

    return result


@given(
    limit=st.integers(min_value=1, max_value=10),
    offset=st.integers(min_value=0, max_value=10),
    tenants_and_databases=tenants_and_databases_st(max_tenants=10, max_databases=10),
)
def test_list_databases_with_limit_offset(
    limit: int,
    offset: int,
    tenants_and_databases: Dict[str, List[str]],
    client_factories: ClientFactories,
) -> None:
    client = client_factories.create_client()
    client.reset()

    admin_client = client_factories.create_admin_client_from_system()

    for tenant, databases in tenants_and_databases.items():
        admin_client.create_tenant(tenant)

        for database in databases:
            admin_client.create_database(database, tenant)

    for tenant, all_databases in tenants_and_databases.items():
        listed_databases = admin_client.list_databases(
            limit=limit, offset=offset, tenant=tenant
        )
        expected_databases = all_databases[offset : offset + limit]

        if limit + offset > len(all_databases):
            assert len(listed_databases) == max(len(all_databases) - offset, 0)
            assert [d["name"] for d in listed_databases] == expected_databases
        else:
            assert len(listed_databases) == limit
            assert [d["name"] for d in listed_databases] == expected_databases



import pytest
from typing import List, Any, Callable, Dict
from jsonschema import ValidationError
from unittest.mock import MagicMock, create_autospec
from chromadb.utils.embedding_functions.schemas import (
    validate_config_schema,
    load_schema,
    get_available_schemas,
)
from chromadb.utils.embedding_functions import (
    known_embedding_functions,
    sparse_known_embedding_functions,
)
from chromadb.api.types import Documents, Embeddings
from pytest import MonkeyPatch

# Skip these embedding functions in tests
SKIP_EMBEDDING_FUNCTIONS = [
    "chroma_langchain",
]


def get_embedding_function_names() -> List[str]:
    """Get all embedding function names to test"""
    return [
        name
        for name in known_embedding_functions.keys()
        if name not in SKIP_EMBEDDING_FUNCTIONS
    ]


class TestEmbeddingFunctionSchemas:
    """Test class for embedding function schemas"""

    @pytest.mark.parametrize("ef_name", get_embedding_function_names())
    def test_embedding_function_config_roundtrip(
        self,
        ef_name: str,
        mock_embeddings: Callable[[Documents], Embeddings],
        mock_common_deps: MonkeyPatch,
    ) -> None:
        """Test embedding function configuration roundtrip"""
        ef_class = known_embedding_functions[ef_name]

        # Create an autospec of the embedding function class
        mock_ef = create_autospec(ef_class, instance=True)

        # Mock the __call__ method
        mock_call = MagicMock(return_value=mock_embeddings(["test"]))
        mock_ef.__call__ = mock_call

        # For chroma-cloud-qwen, mock get_config to return valid data
        if ef_name == "chroma-cloud-qwen":
            from chromadb.utils.embedding_functions.chroma_cloud_qwen_embedding_function import (
                ChromaCloudQwenEmbeddingModel,
                CHROMA_CLOUD_QWEN_DEFAULT_INSTRUCTIONS,
            )

            mock_ef.get_config.return_value = {
                "api_key_env_var": "CHROMA_API_KEY",
                "model": ChromaCloudQwenEmbeddingModel.QWEN3_EMBEDDING_0p6B.value,
                "task": "nl_to_code",
                "instructions": CHROMA_CLOUD_QWEN_DEFAULT_INSTRUCTIONS,
            }

        # Mock the class constructor to return our mock instance
        mock_common_deps.setattr(
            ef_class, "__new__", lambda cls, *args, **kwargs: mock_ef
        )

        # Create instance with minimal args (constructor will be mocked)
        ef_instance = ef_class()

        # Get the config (this will use the real method)
        config = ef_instance.get_config()

        # Test recreation from config
        new_instance = ef_class.build_from_config(config)
        new_config = new_instance.get_config()

        # Configs should match
        assert (
            config == new_config
        ), f"Configs don't match after recreation for {ef_name}"

    def test_schema_required_fields(self) -> None:
        """Test that schemas enforce required fields"""
        for schema_name in get_available_schemas():
            schema = load_schema(schema_name)
            if "required" not in schema:
                continue

            # Create minimal valid config
            config = {}
            for field in schema["required"]:
                field_schema = schema["properties"][field]
                field_type = (
                    field_schema["type"][0]
                    if isinstance(field_schema["type"], list)
                    else field_schema["type"]
                )
                config[field] = self._get_dummy_value(field_type)

            # Test each required field
            for field in schema["required"]:
                test_config = config.copy()
                del test_config[field]
                with pytest.raises(ValidationError):
                    validate_config_schema(test_config, schema_name)

    @staticmethod
    def _get_dummy_value(field_type: str) -> Any:
        """Get a dummy value for a given field type"""
        type_map = {
            "string": "dummy",
            "integer": 0,
            "number": 0.0,
            "boolean": False,
            "object": {},
            "array": [],
        }
        return type_map.get(field_type, "dummy")

    def test_schema_additional_properties(self) -> None:
        """Test that schemas reject additional properties"""
        for schema_name in get_available_schemas():
            schema = load_schema(schema_name)
            config = {}

            # Add required fields
            if "required" in schema:
                for field in schema["required"]:
                    field_schema = schema["properties"][field]
                    field_type = (
                        field_schema["type"][0]
                        if isinstance(field_schema["type"], list)
                        else field_schema["type"]
                    )
                    config[field] = self._get_dummy_value(field_type)

            # Add additional property
            test_config = config.copy()
            test_config["additional_property"] = "value"

            # Test validation
            if schema.get("additionalProperties", True) is False:
                with pytest.raises(ValidationError):
                    validate_config_schema(test_config, schema_name)

    def _create_valid_config_from_schema(
        self, schema: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Create a valid config from a schema by filling in required fields"""
        config: Dict[str, Any] = {}

        if "required" in schema and "properties" in schema:
            for field in schema["required"]:
                if field in schema["properties"]:
                    field_schema = schema["properties"][field]
                    config[field] = self._get_value_from_field_schema(field_schema)

        return config

    def _get_value_from_field_schema(self, field_schema: Dict[str, Any]) -> Any:
        """Get a valid value from a field schema"""
        # Handle enums - use first enum value
        if "enum" in field_schema:
            return field_schema["enum"][0]

        # Handle type (could be a list or single value)
        field_type = field_schema.get("type")
        if field_type is None:
            return "dummy"  # Fallback if no type specified

        if isinstance(field_type, list):
            # If null is in the type list, prefer non-null type
            non_null_types = [t for t in field_type if t != "null"]
            field_type = non_null_types[0] if non_null_types else field_type[0]

        if field_type == "object":
            # Handle nested objects
            nested_config = {}
            if "properties" in field_schema:
                nested_required = field_schema.get("required", [])
                for prop in nested_required:
                    if prop in field_schema["properties"]:
                        nested_config[prop] = self._get_value_from_field_schema(
                            field_schema["properties"][prop]
                        )
            return nested_config if nested_config else {}

        if field_type == "array":
            # Return empty array for arrays
            return []

        # Use the existing dummy value method for primitive types
        return self._get_dummy_value(field_type)

    def _has_custom_validation(self, ef_class: Any) -> bool:
        """Check if validate_config actually validates (not just base implementation)"""
        try:
            # Try with an obviously invalid config - if it doesn't raise, it's base implementation
            invalid_config = {"__invalid_test_config__": True}
            try:
                ef_class.validate_config(invalid_config)
                # If we get here without exception, it's using base implementation
                return False
            except (ValidationError, ValueError, FileNotFoundError):
                # If it raises any validation-related error, it's actually validating
                return True
        except Exception:
            # Any other exception means it's trying to validate (e.g., schema not found)
            return True

    def _setup_env_vars_for_ef(
        self, ef_name: str, mock_common_deps: MonkeyPatch
    ) -> None:
        """Set up environment variables needed for embedding function instantiation"""
        # Map of embedding function names to their default API key environment variable names
        api_key_env_vars = {
            "cohere": "CHROMA_COHERE_API_KEY",
            "openai": "CHROMA_OPENAI_API_KEY",
            "huggingface": "CHROMA_HUGGINGFACE_API_KEY",
            "huggingface_server": "CHROMA_HUGGINGFACE_API_KEY",
            "google_palm": "CHROMA_GOOGLE_PALM_API_KEY",
            "google_generative_ai": "CHROMA_GOOGLE_GENAI_API_KEY",
            "google_vertex": "CHROMA_GOOGLE_VERTEX_API_KEY",
            "jina": "CHROMA_JINA_API_KEY",
            "mistral": "MISTRAL_API_KEY",
            "morph": "MORPH_API_KEY",
            "voyageai": "CHROMA_VOYAGE_API_KEY",
            "cloudflare_workers_ai": "CHROMA_CLOUDFLARE_API_KEY",
            "together_ai": "CHROMA_TOGETHER_AI_API_KEY",
            "baseten": "CHROMA_BASETEN_API_KEY",
            "roboflow": "CHROMA_ROBOFLOW_API_KEY",
            "amazon_bedrock": "AWS_ACCESS_KEY_ID",  # AWS uses different env vars
            "chroma-cloud-qwen": "CHROMA_API_KEY",
            # Sparse embedding functions
            "chroma-cloud-splade": "CHROMA_API_KEY",
        }

        # Set API key environment variable if needed
        if ef_name in api_key_env_vars:
            mock_common_deps.setenv(api_key_env_vars[ef_name], "test-api-key")

        # Special cases that need additional environment variables
        if ef_name == "amazon_bedrock":
            mock_common_deps.setenv("AWS_SECRET_ACCESS_KEY", "test-secret-key")
            mock_common_deps.setenv("AWS_REGION", "us-east-1")

    def _create_ef_instance(
        self, ef_name: str, ef_class: Any, mock_common_deps: MonkeyPatch
    ) -> Any:
        """Create an embedding function instance, handling special cases"""
        # Set up environment variables first
        self._setup_env_vars_for_ef(ef_name, mock_common_deps)

        # Mock missing modules that are imported inside __init__ methods
        import sys

        # Create mock modules
        mock_pil = MagicMock()
        mock_pil_image = MagicMock()
        mock_google_genai = MagicMock()
        mock_vertexai = MagicMock()
        mock_vertexai_lm = MagicMock()
        mock_boto3 = MagicMock()
        mock_jina = MagicMock()
        mock_mistralai = MagicMock()

        # Mock boto3.Session for amazon_bedrock
        mock_boto3_session = MagicMock()
        mock_session_instance = MagicMock()
        mock_session_instance.region_name = "us-east-1"
        mock_session_instance.profile_name = None
        mock_session_instance.client.return_value = MagicMock()
        mock_boto3_session.return_value = mock_session_instance
        mock_boto3.Session = mock_boto3_session

        # Mock vertexai.init and TextEmbeddingModel
        mock_text_embedding_model = MagicMock()
        mock_text_embedding_model.from_pretrained.return_value = MagicMock()
        mock_vertexai_lm.TextEmbeddingModel = mock_text_embedding_model
        mock_vertexai.language_models = mock_vertexai_lm
        mock_vertexai.init = MagicMock()

        # Mock google.generativeai - need to set up google module first
        mock_google = MagicMock()
        mock_google_genai.configure = MagicMock()  # For palm.configure()
        mock_google_genai.GenerativeModel = MagicMock(return_value=MagicMock())
        mock_google.generativeai = mock_google_genai

        # Mock jina Client
        mock_jina.Client = MagicMock()

        # Mock mistralai
        mock_mistral_client = MagicMock()
        mock_mistral_client.return_value.embeddings.create.return_value.data = [
            MagicMock(embedding=[0.1, 0.2, 0.3])
        ]
        mock_mistralai.Mistral = mock_mistral_client

        # Add missing modules to sys.modules using monkeypatch
        modules_to_mock = {
            "PIL": mock_pil,
            "PIL.Image": mock_pil_image,
            "google": mock_google,
            "google.generativeai": mock_google_genai,
            "vertexai": mock_vertexai,
            "vertexai.language_models": mock_vertexai_lm,
            "boto3": mock_boto3,
            "jina": mock_jina,
            "mistralai": mock_mistralai,
        }

        for module_name, mock_module in modules_to_mock.items():
            mock_common_deps.setitem(sys.modules, module_name, mock_module)

        # Special cases that need additional arguments
        if ef_name == "cloudflare_workers_ai":
            return ef_class(
                model_name="test-model",
                account_id="test-account-id",
            )
        elif ef_name == "baseten":
            # Baseten needs api_key explicitly passed even with env var
            return ef_class(
                api_key="test-api-key",
                api_base="https://test.api.baseten.co",
            )
        elif ef_name == "amazon_bedrock":
            # Amazon Bedrock needs a boto3 session - create a mock session
            # boto3 is already mocked in sys.modules above
            mock_session = mock_boto3.Session(region_name="us-east-1")
            return ef_class(
                session=mock_session,
                model_name="amazon.titan-embed-text-v1",
            )
        elif ef_name == "huggingface_server":
            return ef_class(url="http://localhost:8080")
        elif ef_name == "google_vertex":
            return ef_class(project_id="test-project", region="us-central1")
        elif ef_name == "mistral":
            return ef_class(model="mistral-embed")
        elif ef_name == "roboflow":
            return ef_class()  # No model_name needed
        elif ef_name == "chroma-cloud-qwen":
            from chromadb.utils.embedding_functions.chroma_cloud_qwen_embedding_function import (
                ChromaCloudQwenEmbeddingModel,
            )

            return ef_class(
                model=ChromaCloudQwenEmbeddingModel.QWEN3_EMBEDDING_0p6B,
                task="nl_to_code",
            )
        else:
            # Try with no args first
            try:
                return ef_class()
            except Exception:
                # If that fails, try with common minimal args
                return ef_class(model_name="test-model")

    @pytest.mark.parametrize("ef_name", get_embedding_function_names())
    def test_validate_config_with_schema(
        self,
        ef_name: str,
        mock_embeddings: Callable[[Documents], Embeddings],
        mock_common_deps: MonkeyPatch,
    ) -> None:
        """Test that validate_config works correctly with actual configs from embedding functions"""
        ef_class = known_embedding_functions[ef_name]

        # Skip if the embedding function doesn't have a validate_config method
        if not hasattr(ef_class, "validate_config"):
            pytest.skip(f"{ef_name} does not have validate_config method")

        # Check if it's callable (static methods are callable on the class)
        if not callable(getattr(ef_class, "validate_config", None)):
            pytest.skip(f"{ef_name} validate_config is not callable")

        # Skip if using base implementation (doesn't actually validate)
        if not self._has_custom_validation(ef_class):
            pytest.skip(
                f"{ef_name} uses base validate_config implementation (no validation)"
            )

        # Create a real instance to get the actual config
        # We'll mock __call__ to avoid needing to actually generate embeddings
        try:
            ef_instance = self._create_ef_instance(ef_name, ef_class, mock_common_deps)
        except Exception as e:
            pytest.skip(
                f"{ef_name} requires arguments that we cannot provide without external deps: {e}"
            )

        # Mock only __call__ to avoid needing to actually generate embeddings
        mock_call = MagicMock(return_value=mock_embeddings(["test"]))
        mock_common_deps.setattr(ef_instance, "__call__", mock_call)

        # Get the actual config from the embedding function (this uses the real get_config method)
        config = ef_instance.get_config()

        # Filter out None values - optional fields with None shouldn't be included in validation
        # This matches common JSON schema practice where optional fields are omitted rather than null
        config = {k: v for k, v in config.items() if v is not None}

        # Validate the actual config using the embedding function's validate_config method
        ef_class.validate_config(config)

    def test_validate_config_sparse_embedding_functions(
        self,
        mock_embeddings: Callable[[Documents], Embeddings],
        mock_common_deps: MonkeyPatch,
    ) -> None:
        """Test validate_config for sparse embedding functions with actual configs"""
        for ef_name, ef_class in sparse_known_embedding_functions.items():
            # Skip if the embedding function doesn't have a validate_config method
            if not hasattr(ef_class, "validate_config"):
                continue

            # Check if it's callable (static methods are callable on the class)
            if not callable(getattr(ef_class, "validate_config", None)):
                continue

            # Skip if using base implementation (doesn't actually validate)
            if not self._has_custom_validation(ef_class):
                continue

            # Create a real instance to get the actual config
            # We'll mock __call__ to avoid needing to actually generate embeddings
            try:
                ef_instance = self._create_ef_instance(
                    ef_name, ef_class, mock_common_deps
                )
            except Exception:
                continue  # Skip if we can't create instance

            # Mock only __call__ to avoid needing to actually generate embeddings
            mock_call = MagicMock(return_value=mock_embeddings(["test"]))
            mock_common_deps.setattr(ef_instance, "__call__", mock_call)

            # Get the actual config from the embedding function (this uses the real get_config method)
            config = ef_instance.get_config()

            # Filter out None values - optional fields with None shouldn't be included in validation
            # This matches common JSON schema practice where optional fields are omitted rather than null
            config = {k: v for k, v in config.items() if v is not None}

            # Validate the actual config using the embedding function's validate_config method
            ef_class.validate_config(config)



from hypothesis import given, strategies as st
from chromadb.api.types import (
    optional_embeddings_to_base64_strings,
    optional_base64_strings_to_embeddings,
)
import numpy as np
import math


@given(st.lists(st.lists(st.integers(min_value=-128, max_value=127))))
def test_base64_conversion_is_identity_i8(embeddings) -> None:  # type: ignore
    b64_strings = optional_embeddings_to_base64_strings(embeddings)
    assert b64_strings is not None
    assert len(b64_strings) == len(embeddings)
    decoded_embeddings = optional_base64_strings_to_embeddings(b64_strings)
    for orig, decoded in zip(embeddings, decoded_embeddings):  # type: ignore
        np.testing.assert_allclose(orig, decoded, rtol=1e-6)


@given(st.lists(st.lists(st.floats(width=16))))
def test_base64_conversion_is_identity_f16(embeddings) -> None:  # type: ignore
    b64_strings = optional_embeddings_to_base64_strings(embeddings)
    assert b64_strings is not None
    assert len(b64_strings) == len(embeddings)
    decoded_embeddings = optional_base64_strings_to_embeddings(b64_strings)
    for orig, decoded in zip(embeddings, decoded_embeddings):  # type: ignore
        np.testing.assert_allclose(orig, decoded, rtol=1e-6)


@given(st.lists(st.lists(st.floats(width=32))))
def test_base64_conversion_is_identity_f32(embeddings) -> None:  # type: ignore
    b64_strings = optional_embeddings_to_base64_strings(embeddings)
    assert b64_strings is not None
    assert len(b64_strings) == len(embeddings)
    decoded_embeddings = optional_base64_strings_to_embeddings(b64_strings)
    for orig, decoded in zip(embeddings, decoded_embeddings):  # type: ignore
        np.testing.assert_allclose(orig, decoded, rtol=1e-6)


@given(st.lists(st.lists(st.floats(width=64))))
def test_base64_conversion_is_identity_f64(embeddings) -> None:  # type: ignore
    b64_strings = optional_embeddings_to_base64_strings(embeddings)
    assert b64_strings is not None
    assert len(b64_strings) == len(embeddings)
    decoded_embeddings = optional_base64_strings_to_embeddings(b64_strings)

    expected_embeddings = []
    for embedding in embeddings:
        expected_embedding = []
        for value in embedding:
            if math.isnan(value):
                expected_embedding.append(float("nan"))
            elif value > np.finfo(np.float32).max:
                expected_embedding.append(float("inf"))
            elif value < np.finfo(np.float32).min:
                expected_embedding.append(float("-inf"))
            else:
                f32_value = np.float32(value)
                expected_embedding.append(float(f32_value))
        expected_embeddings.append(expected_embedding)

    for orig, decoded in zip(expected_embeddings, decoded_embeddings):  # type: ignore
        np.testing.assert_allclose(orig, decoded, rtol=1e-6)


@given(st.lists(st.lists(st.floats(width=32))))
def test_base64_conversion_numpy_is_identity_f32(embeddings) -> None:  # type: ignore
    b64_strings = optional_embeddings_to_base64_strings(
        [np.array(embedding, dtype=np.float32) for embedding in embeddings]
    )
    assert b64_strings is not None
    assert len(b64_strings) == len(embeddings)
    decoded_embeddings = optional_base64_strings_to_embeddings(b64_strings)

    expected_embeddings = []
    for embedding in embeddings:
        expected_embedding = []
        for value in embedding:
            if math.isnan(value):
                expected_embedding.append(float("nan"))
            elif value > np.finfo(np.float32).max:
                expected_embedding.append(float("inf"))
            elif value < np.finfo(np.float32).min:
                expected_embedding.append(float("-inf"))
            else:
                f32_value = np.float32(value)
                expected_embedding.append(float(f32_value))
        expected_embeddings.append(expected_embedding)

    for orig, decoded in zip(expected_embeddings, decoded_embeddings):  # type: ignore
        np.testing.assert_allclose(orig, decoded, rtol=1e-6)
